"""
Sub-module with melt-only thermometers
"""

import inspect
import sys

import elementMass as e
import numpy as np
import pandas as pd

from MagmaPandas.magma_protocol import Magma
from MagmaPandas.parse_io import check_components
from MagmaPandas.thermometers.validate_temperatures import _check_temperature
from MagmaPandas.tools.calibration_range import _check_calibration_range
from MagmaPandas.tools.modify_compositions import (
    _anhydrous_composition,
    _get_elements,
    _remove_elements,
    moles_per_oxygen,
)

calibration_range = {
    "putirka2008_14": [
        ["SiO2", 31, 73.64],
        [["Na2O", "K2O"], 0, 14.3],
        ["H2O", 0, 18.6],
    ],
    "putirka2008_15": [
        ["SiO2", 31, 73.64],
        [["Na2O", "K2O"], 0, 14.3],
        ["H2O", 0, 18.6],
    ],
}

errors = pd.Series(
    {
        "putirka2008_13": 71,
        "putirka2008_14": 58,
        "putirka2008_15": 46,
        "putirka2008_16": 26,
        "putirka2008_22": 32,  # from Wieser et al. (2025) Treatise of Geochemistry
        "sun2020": 49,
        "shea2022": 13,  # on calibration dataset, not validated
        "sugawara2000_3": 33,  # on calibration dataset
        "sugawara2000_6a": 30,  # on calibration dataset
    }
)

components = {
    "putirka2008_13": ["MgO"],
    "putirka2008_14": ["MgO", "FeO", "Na2O", "K2O", "H2O"],
    "putirka2008_15": ["MgO", "FeO", "Na2O", "K2O", "H2O"],
    "putirka2008_16": ["SiO2", "Al2O3", "MgO"],
    "putirka2008_22": ["SiO2", "FeO", "MnO", "MgO", "CaO", "CoO", "NiO", "H2O"],
    "sun2020": ["MgO", "CaO", "K2O", "TiO2", "FeO", "CO2", "H2O"],
    "shea2022": ["MgO"],
    "sugawara2000_3": ["MgO"],
    "sugawara2000_6a": ["MgO", "FeO", "CaO", "SiO2"],
}

_Beattie_constants = pd.DataFrame(
    {
        "A": [1.0, 0.259, 0.299, 0.786, 3.346],
        "B": [0.0, -4.9e-2, 2.7e-2, -0.385, -3.665],
    },
    index=["Mg", "Mn", "Fe", "Co", "Ni"],
)


def putirka2008_13(
    melt: Magma, offset: float = 0.0, *args, **kwargs
) -> float | pd.Series:
    """
    melt-only thermometer

    Equation 13 from Putirka (2008)\ [28]_ calculates liquidus temperatures based on melt compositions.
    Requires saturation in olivine.

    SEE = 71 degrees

    Parameters
    ----------
    melt : Magma
        melt compositions in oxide wt. %

    offset : float
        offset value in standard deviations. Temperatures are calculated as ``temnperature + offset * thermometer error (SEE)``.

    Returns
    -------
    temperatures : float, pandas Series
        liquidus temperatures in Kelvin.
    """

    composition = check_components(
        composition=melt, components=components["putirka2008_13"]
    )

    T_K = 26.3 * composition["MgO"] + 994.4 + 273.15

    _check_temperature(T_K)

    T_K = T_K + errors["putirka2008_13"] * offset

    return pd.Series(T_K, name="T_K").squeeze()


def putirka2008_14(
    melt: Magma, warn=True, offset: float = 0.0, *args, **kwargs
) -> float | pd.Series:
    """
    melt-only thermometer

    Equation 14 from Putirka (2008)\ [28]_ calculates liquidus temperatures based on melt compositions.
    Requires saturation in olivine.

    SEE = 58 degrees

    Applicable between:

    -    0 - 14.4 GPa

    -    729 - 2000 degrees C

    -    31 - 73.64 wt. % SiO\ :sub:`2`

    -    0 - 14.3 wt. % Na\ :sub:`2`\ O + K\ :sub:`2`\ O

    -    0 - 18.6 wt. % H\ :sub:`2`\ O

    Parameters
    ----------
    melt : Magma
        melt compositions in oxide wt. %

    offset : float
        offset value in standard deviations. Temperatures are calculated as ``temnperature + offset * thermometer error (SEE)``.

    Returns
    -------
    temperatures : float, pandas Series
        liquidus temperatures in Kelvin.
    """

    elements = _get_elements(melt)
    composition = check_components(
        composition=melt, components=components["putirka2008_14"]
    )

    if warn:
        _check_calibration_range(
            melt=composition, calibration_range=calibration_range["putirka2008_14"]
        )
    # oxides_needed = set(["MgO", "FeO", "Na2O", "K2O"])

    # composition = parse_data(melt, oxides_needed)

    H2O = composition["H2O"].copy()
    composition = _anhydrous_composition(composition)

    # Calculate molar oxide fractions
    mol_fractions = composition.moles()
    # Melt Mg#
    Mg_no = mol_fractions["MgO"] / (
        mol_fractions["MgO"] + mol_fractions["FeO"]
    )  # Putirka doesn't specify whether this should be Fe2+ or Fe(total)

    if "Fe2O3" in elements:
        FeO_w, Fe2O3_w = e.compound_weights(["FeO", "Fe2O3"])
        composition["FeO"] = composition["FeO"] + (
            2 * composition["Fe2O3"] * (FeO_w / Fe2O3_w)
        )

    T_K = (
        754
        + 190.6 * Mg_no
        + 25.52 * composition["MgO"]
        + 9.585 * composition["FeO"]
        + 14.87 * (composition["Na2O"] + composition["K2O"])
        - 9.176 * H2O
    ) + 273.15

    _check_temperature(T_K)

    T_K = T_K + errors["putirka2008_14"] * offset

    return pd.Series(T_K, name="T_K").squeeze()


def putirka2008_15(
    melt: Magma,
    P_bar: float | pd.Series,
    offset: float = 0.0,
    warn=True,
    **kwargs,
) -> float | pd.Series:
    """
    melt-only thermometer

    Equation 15 from Putirka (2008)\ [28]_ calculates liquidus temperatures based on melt compositions.
    Requires saturation in olivine.

    SEE = 46 degrees

    Applicable between:

    -    0 - 14.4 GPa

    -    729 - 2000 degrees C

    -    31 - 73.64 wt. % SiO\ :sub:`2`

    -    0 - 14.3 wt. % Na\ :sub:`2`\ O + K\ :sub:`2`\ O

    -    0 - 18.6 wt. % H\ :sub:`2`\ O

    Parameters
    ----------
    melt : Magma
        melt compositions in oxide wt. %

    P_bar : float, pandas Series
        pressures in bar.

    offset : float
        offset value in standard deviations. Temperatures are calculated as ``temnperature + offset * thermometer error (SEE)``.

    Returns
    -------
    temperatures : float, pandas Series
        liquidus temperatures in Kelvin.
    """

    elements = _get_elements(melt)
    composition = check_components(
        composition=melt, components=components["putirka2008_15"]
    )

    if warn:
        _check_calibration_range(
            melt=composition, calibration_range=calibration_range["putirka2008_15"]
        )
    # oxides_needed = set(["MgO", "FeO", "Na2O", "K2O"])
    # composition = parse_data(melt, oxides_needed)

    H2O = composition["H2O"].copy()
    composition = _anhydrous_composition(composition)

    P_GPa = P_bar / 1e4

    # Calculate molar oxide fractions
    mol_fractions = composition.moles()
    # Melt Mg#
    Mg_no = mol_fractions["MgO"] / (
        mol_fractions["MgO"] + mol_fractions["FeO"]
    )  # Putirka doesn't specify whether this should be Fe2+ or Fe(total)

    if "Fe2O3" in elements:
        FeO_w, Fe2O3_w = e.compound_weights(["FeO", "Fe2O3"])
        composition["FeO"] = composition["FeO"] + (
            2 * composition["Fe2O3"] * (FeO_w / Fe2O3_w)
        )

    T_K = (
        815.3
        + 265.5 * Mg_no
        + 15.37 * composition["MgO"]
        + 8.61 * composition["FeO"]
        + 6.646 * (composition["Na2O"] + composition["K2O"])
        + 39.16 * P_GPa
        - 12.83 * H2O
    ) + 273.15

    _check_temperature(T_K)

    T_K = T_K + errors["putirka2008_15"] * offset

    return pd.Series(T_K, name="T_K").squeeze()


def putirka2008_16(
    melt: Magma, P_bar: float | pd.Series, offset: float = 0.0, **kwargs
) -> float | pd.Series:
    """
    melt-only thermometer

    Equation 16 from Putirka (2008)\ [28]_ calculates liquiqdus temperatures based on melt compositions.
    Requires equilibrium with olivine + plagioclase + clinopyroxene and saturation with additional phases drastically increases the standard error of estimate.

    SEE = 26 degrees (saturation in olivine + plagioclase + clinopyroxene)
        = 60 degrees (saturation with additional phases)

    Parameters
    ----------
    melt : Magma
        melt compositions in oxide wt. %

    P_bar : float, pandas Series
        pressures in bar

    offset : float
        offset value in standard deviations. Temperatures are calculated as ``temnperature + offset * thermometer error (SEE)``.

    Returns
    -------
    temperatures : float, pandas Series
        liquidus temperatures in Kelvin.
    """
    elements = _get_elements(melt)
    composition = check_components(
        composition=melt, components=components["putirka2008_16"]
    )
    # oxides_needed = set(["SiO2", "Al2O3", "MgO"])

    # composition = parse_data(melt, oxides_needed)
    # import MagmaPandas as mp

    if isinstance(P_bar, pd.Series):
        if not melt.index.equals(P_bar.index):
            raise RuntimeError("Melt and P_bar indices don't match")

    if "H2O" in elements:
        composition = _anhydrous_composition(composition)

    # Convert pressure from bars to GPa
    P_GPa = P_bar / 1e4

    # Calculate molar oxide fractions
    mol_fractions = composition.moles()

    part_1 = (
        -583
        + 3141 * mol_fractions["SiO2"]
        + 15779 * mol_fractions["Al2O3"]
        + 1338.6 * mol_fractions["MgO"]
    )
    part_2 = -31440 * mol_fractions["SiO2"] * mol_fractions["Al2O3"] + 77.67 * P_GPa

    T_K = part_1 + part_2 + 273.15

    _check_temperature(T_K)

    T_K = T_K + errors["putirka2008_16"] * offset

    return pd.Series(T_K, name="T_K").squeeze()


def putirka2008_22(
    melt: Magma,
    P_bar: float | pd.Series,
    offset: float = 0.0,
    **kwargs,
) -> float | pd.Series:
    """
    melt-only thermometer

    Equation 22 from Putirka (2008)\ [28]_, combined with equation 12 from Beattie (1993) calculates liquidus temperatures based on melt compositions. #TODO add reference link


    Parameters
    ----------
    melt : Magma
        melt compositions in oxide wt. %

    P_bar : float, pandas Series
        pressures in bar.

    offset : float
        offset value in standard deviations. Temperatures are calculated as ``temnperature + offset * thermometer error (SEE)``.

    Returns
    -------
    temperatures : float, pandas Series
        liquidus temperatures in Kelvin.
    """

    axis = [0, 1][isinstance(melt, pd.DataFrame)]

    composition = check_components(
        composition=melt, components=components["putirka2008_22"]
    )
    H2O = composition["H2O"].copy()
    composition = _anhydrous_composition(composition)

    # if warn:
    #     _check_calibration_range(
    #         melt=composition, calibration_range=calibration_range["putirka2008_22"]
    # )

    P_GPa = P_bar / 1e4

    cations = composition.cations()
    lnD_Mg = np.log(
        (
            0.666
            - cations[_Beattie_constants.index]
            .mul(_Beattie_constants.loc[:, "B"], axis=axis)
            .sum(axis=axis)
        )
        / cations[_Beattie_constants.index]
        .mul(_Beattie_constants.loc[:, "A"], axis=axis)
        .sum(axis=axis)
    )  # equation 12 from Beattie (1993)

    C_NM = (cations[["Fe", "Mn", "Mg", "Ca", "Co", "Ni"]]).sum(axis=axis)
    NF = 7 / 2 * np.log(1 - cations["Al"]) + 7 * np.log(
        1 - cations["Ti"]
    )  # Table 1 from Putirka (2008)

    T_K = (15294.6 + 1318.8 * P_GPa + 2.4834 * P_GPa**2) / (
        8.048
        + 2.8352 * lnD_Mg
        + 2.097 * np.log(1.5 * C_NM)
        + 2.575 * np.log(3 * cations["Si"])
        - 1.41 * NF
        + 0.222 * H2O
        + 0.5 * P_GPa
    ) + 273.15  # equation 22 from Putirka (2008)

    _check_temperature(T_K)

    T_K = T_K + errors["putirka2008_22"] * offset

    return pd.Series(T_K, name="T_K").squeeze()


def sun2020(melt, P_bar: float | pd.Series, offset: float = 0.0, **kwargs):
    """
    Equation 6 from:

    Sun and Dasgupta (2020)\ [22]_

    Calibrated at:
    ~ 2 - 10 GPa
    ~ 950 - 1600 degrees C

    SEE: 49 degrees C

    Parameters
    ----------
    melt : Magma
        melt compositions in oxide wt. %
    P_bar   :
        pressures in bar
    offset : float
        offset value in standard deviations. Temperatures are calculated as ``temnperature + offset * thermometer error (SEE)``.

    Returns
    -------
    temperatures : float, pandas Series
        liquidus temperatures in Kelvin.
    """

    P_GPa = P_bar / 1e4

    composition = check_components(composition=melt, components=components["sun2020"])
    moles = composition.moles()

    composition_volatile_free = _remove_elements(
        composition=moles, drop=["H2O", "CO2", "F", "S", "Cl"]
    )
    moles_unit_oxygen = moles_per_oxygen(moles=composition_volatile_free)

    Omega = (
        2.59
        + 3.5 * (moles_unit_oxygen["Ca1O"] - 2 * moles_unit_oxygen["K2O"])
        + 4.85 * moles_unit_oxygen["Ti1/2O"]
        + 1.4
        * (
            moles_unit_oxygen["Mg1O"]
            / (moles_unit_oxygen["Mg1O"] + moles_unit_oxygen["Fe1O"])
        )
        + 0.5 * moles_unit_oxygen["Mg1O"] * np.sqrt(composition["CO2"])
        + 5.7e-2 * composition["H2O"]
    )

    T_K = 1e4 / (
        Omega - 0.34 * np.sqrt(P_GPa) - 1.26 * np.log(moles_unit_oxygen["Mg1O"])
    )

    _check_temperature(T_K)

    T_K = T_K + errors["sun2020"] * offset

    return pd.Series(T_K, name="T_K").squeeze()


def shea2022(melt, offset: float = 0.0, **kwargs):
    """
    Equation 1 from Shea et al. (2022)\ [30]_

    Calibrated at:
    1 bar
    1060 - 1500 degrees C

    SEE: 13 degrees C (on calibration dataset, not validated)

    Parameters
    ----------
    melt : Magma
        melt compositions in oxide wt. %

    offset : float
        offset value in standard deviations. Temperatures are calculated as ``temnperature + offset * thermometer error (SEE)``.

    Returns
    -------
    temperatures : float, pandas Series
        liquidus temperatures in Kelvin.
    """

    composition = check_components(composition=melt, components=components["shea2022"])

    T_K = 21.2 * composition["MgO"] + 1017 + 273.15

    _check_temperature(T_K)

    T_K = T_K + errors["shea2022"] * offset

    return pd.Series(T_K, name="T_K").squeeze()


def sugawara2000_3(melt, P_bar: float | pd.Series, offset: float = 0.0, **kwargs):
    """
    Equation 3 with olivine-liquid parameters and corrected for H2O according to equation 7a from Sugawara (2000)\ [31]_

    Calibrated at:
    <= 3.5 GPa
    1266 - 1873 C

    SEE: 33 degrees C

    Parameters
    ----------
    melt : Magma
        melt compositions in oxide wt. %
    P_bar   :
        pressures in bar
    offset : float
        offset value in standard deviations. Temperatures are calculated as ``temnperature + offset * thermometer error (SEE)``.

    Returns
    -------
    temperatures : float, pandas Series
        liquidus temperatures in Kelvin.
    """

    composition = check_components(
        composition=melt, components=components["sugawara2000_3"]
    )
    moles_percent = composition.moles() * 100

    A = 1293
    B = 14.60
    C = 5.5e-3
    T_K = A + B * moles_percent["MgO"] + C * P_bar

    if "H2O" in moles_percent.elements:
        T_K = T_K - 5.403 * moles_percent["H2O"]

    _check_temperature(T_K)

    T_K = T_K + errors["sugawara2000_3"] * offset

    return pd.Series(T_K, name="T_K").squeeze()


def sugawara2000_6a(melt, P_bar: float | pd.Series, offset: float = 0.0, **kwargs):
    """
    Equation 6a corrected for H2O according to equation 7a from Sugawara (2000)\ [31]_
    Calibrated at:
    <= 3.5 GPa
    1266 - 1873 C

    SEE: 30 degrees C

    Parameters
    ----------
    melt : Magma
        melt compositions in oxide wt. %
    P_bar   :
        pressures in bar
    offset : float
        offset value in standard deviations. Temperatures are calculated as ``temnperature + offset * thermometer error (SEE)``.

    Returns
    -------
    temperatures : float, pandas Series
        liquidus temperatures in Kelvin.
    """

    composition = check_components(
        composition=melt, components=components["sugawara2000_6a"]
    )
    moles_percent = composition.moles() * 100

    T_K = (
        1466
        - 1.44 * moles_percent["SiO2"]
        - 0.5 * moles_percent["FeO"]
        + 12.32 * moles_percent["MgO"]
        - 3.899 * moles_percent["CaO"]
        + 4.3e-3 * P_bar
    )

    if "H2O" in moles_percent.elements:
        T_K = T_K - 5.403 * moles_percent["H2O"]

    _check_temperature(T_K)

    T_K = T_K + errors["sugawara2000_6a"] * offset

    return pd.Series(T_K, name="T_K").squeeze()


_module = sys.modules[__name__]
_funcs = inspect.getmembers(_module, inspect.isfunction)
# collect all melt thermometers in a dictionary
melt_thermometers_dict = {
    f[0]: f[1] for f in _funcs if f[1].__module__ == _module.__name__
}



import warnings as w

import numpy as np


def _check_temperature(T_K):
    """
    Find negative or NaN temperatures
    """
    try:
        if (neg := any(T_K < 0)) or (nan := any(np.isnan(T_K))):
            str_arr = np.array(["negative", "NaN"])[np.array([neg, nan])]
            w.warn(f"{', '.join(str_arr)} temperatures found!")
    except TypeError:
        if (neg := (T_K < 0)) or (nan := (T_K != T_K)):
            str_arr = np.array(["negative", "NaN"])[np.array([neg, nan])]



import numpy as np
import scipy.optimize as opt

from MagmaPandas.EOSs.tools import thermal_expansivity


def Vinet_P(V, V_0, K_0, Kprime_0):
    """
    Calculate pressure at room temperature with Vinet EOS.

    Parameters
    ----------
    V : float, array-like
        volume in cm3/mol
    V_0 : float, array-like
        volume at reference conditions
    K : float, array-like
        bulk modulus in GPa
    Kprime_0
        pressure derivative of K in GPa-1

    Returns
    -------
    P_GPa : float, array-like
        pressure in GPa
    """

    x = (V / V_0) ** (1 / 3)
    P_GPa = 3 * K_0 * x**-2 * (1 - x) * np.exp(1.5 * (Kprime_0 - 1) * (1 - x))
    return P_GPa


def Vinet_thermal(V, V_0, alpha0, delta0, kappa, T_K, Tref):
    """
    Calculate thermal contribution to volume with Vinet EOS.


    Parameters
    ----------
    V : float, array-like
        volume in cm3/mol at reference temperature Tref
    V_0 : float
        volume in cm3/mol at reference conditions
    alpha0 : float
        thermal expansian at 1 bar
    delta0 : float
        Value of Anderson-Grüneisen parameter (delta_T) at 1 bar
    kappa : float
        dimensionless Anderson-Grüneisen parameter.
    T_K  : float, array-like
        temperature in Kelvin
    Tref : float
        reference temperature in Kelvin

    Returns
    -------
    V   : float, array-like
        volume in cm3/mol
    """

    # calculate thermal expansion
    alpha = thermal_expansivity(V=V, V_0=V_0, alpha0=alpha0, delta0=delta0, kappa=kappa)
    # calculate volume at temperature T
    V = V * np.exp(alpha * (T_K - Tref))

    return V


@np.vectorize(excluded=["bracket"])
def Vinet_V_roomtemperature(P_GPa, V_0, K_0, Kprime_0):
    """
    Calculate volume at pressure P_GPa and room temperature with Vinet EOS.

    Parameters
    ----------
    P_GPa   : float, array-like
        pressure in GPa
    V_0 : float, array-like
        volume in cm3/mol at reference conditions
    K_0 : float
        bulk modulus at reference conditions
    Kprime_0
        pressure derivative of K in GPa-1

    Returns
    -------
    V   : float, array-like
        volume in cm3/mol
    """

    Vinet_solve = lambda V: Vinet_P(V=V, V_0=V_0, K_0=K_0, Kprime_0=Kprime_0) - P_GPa

    return opt.fsolve(Vinet_solve, x0=V_0)


def Vinet_V(P_GPa, T_K, V_0, K_0, Kprime_0, delta0, alpha0, kappa, Tref=298.15):
    """
    Calculate volume at pressure P_GPa and temperature T_K with Vinet EOS.

    Parameters
    ----------
    P_GPa : float, array-like
        pressure in GPa
    T_K  : float, array-like
        temperature in Kelvin
    V_0 : float
        volume in cm3/mol at reference conditions
    K_0 : float
        bulk modulus at reference conditions
    Kprime_0 : float
        pressure derivative of K_0
    delta0 : float
        Value of Anderson-Grüneisen parameter (delta_T) at 1 bar
    alpha0 : float
        thermal expansian at 1 bar
    kappa : float
        dimensionless Anderson-Grüneisen parameter.
    Tref : float
        reference temperature in Kelvin

    Returns
    -------
    V   : float, array-like
        volume in cm3/mol
    """

    # calculate volume at pressure P_GPa and room temperature
    V_T0_P = Vinet_V_roomtemperature(P_GPa=P_GPa, V_0=V_0, K_0=K_0, Kprime_0=Kprime_0)

    # calculate thermal part
    V = Vinet_thermal(
        V=V_T0_P, V_0=V_0, alpha0=alpha0, delta0=delta0, kappa=kappa, T_K=T_K, Tref=Tref
    )

    return V


def Vinet_VdP(
    P_GPa, T_K, V_0, K_0, Kprime_0, alpha0, delta0, kappa, n_step=100, **kwargs
):
    """
    Parameters
    ----------
    P_GPa : float, array-like
        pressure in GPa
    T_K  : float, array-like
        temperature in Kelvin
    V_0 : float
        volume at reference conditions
    K_0 : float
        bulk modulus at reference conditions
    Kprime_0 : float
        pressure derivative of K_0
    alpha0 : float
        thermal expansian at 1 bar
    delta0 : float
        Value of Anderson-Grüneisen parameter (delta_T) at 1 bar
    kappa : float
        dimensionless Anderson-Grüneisen parameter.
    n_step : float
        number of steps between 1 bar and P_GPa in the numerical integration of V

    Returns
    -------
    VdP
        volume integrated to from 1 bar to P_GPa in J/mol
    """
    if P_GPa <= 1e-4:
        return 0

    P = np.linspace(start=1e-4, stop=P_GPa, num=n_step)
    V = Vinet_V(
        P_GPa=P,
        T_K=T_K,
        V_0=V_0,
        K_0=K_0,
        Kprime_0=Kprime_0,
        delta0=delta0,
        alpha0=alpha0,
        kappa=kappa,
    )

    VdP = np.trapz(V, P)  # (cm3/mol)*GPa = (1000J/GPa/mol)*GPa

    return VdP * 1000  # convert to J/mol



def fO2_RuRuO2(logshift, T_K, P_bar):
    """
    Armstrong et al. (2018) Supplementary materials equation S3.
    and
    Armstrong et al. (2020)

    calibrated between 1 bar and 25 GPa and 773 - 2773 K

    O'Neill and Nell (1997) have an alternative formulation
    """

    P_GPa = P_bar / 1e4
    offset = 10**logshift

    log10fO2 = (
        (7.782 - 9.96e-3 * P_GPa + 1.932e-3 * P_GPa**2 - 3.76e-5 * P_GPa**3)
        + (-13763 + 592 * P_GPa - 3.955 * P_GPa**2) / T_K
        + (-1.05e6 - 4622 * P_GPa) / T_K**2
    )

    fO2 = 10**log10fO2 * offset

    return fO2



import warnings as w
from typing import Dict

import numpy as np
import pandas as pd

from MagmaPandas.magma_protocol import Magma


def _check_calibration_range_Series(melt: pd.Series, calibration_range: Dict) -> None:

    checks = np.array([], dtype=bool)
    for element, *range in calibration_range:
        row = [element] if isinstance(element, str) else element
        np.append(checks, not (range[0] < melt[row].sum() < range[1]))

    if sum(checks) < 1:
        return

    w.warn(f"sample has composition outside the thermometer calibration range")


def _check_calibration_range_dataframe(melt: Magma, calibration_range: Dict) -> None:

    checks = pd.DataFrame(dtype=bool)

    for element, *range in calibration_range:
        col = [element] if isinstance(element, str) else element
        checks[str(element)] = ~melt[col].sum(axis=1).between(range[0], range[1])

    outside_range = checks.any(axis=1)

    if outside_range.sum() < 1:
        return

    w.warn(
        f"samples {*melt.index[outside_range],} have compositions outside the thermometer calibration range"
    )


def _check_calibration_range(melt: pd.Series | Magma, calibration_range) -> None:

    if isinstance(melt, pd.Series):
        _check_calibration_range_Series(melt, calibration_range)
    elif isinstance(melt, pd.DataFrame):
        _check_calibration_range_dataframe(melt, calibration_range)



from fractions import Fraction
from typing import List

import elementMass as e
import pandas as pd


def _get_elements(composition):

    import MagmaPandas as mp

    if isinstance(composition, mp.MagmaFrame):
        return composition.columns

    if isinstance(composition, mp.MagmaSeries):
        return composition.index


def _remove_elements(composition, drop: List[str]):

    composition_new = composition.copy()

    for e in drop:

        if e in composition_new.index:
            composition_new = composition_new.drop(e)
        try:
            if e in composition_new.columns:
                composition_new = composition_new.drop(columns=[e])
        except AttributeError:
            pass

        composition_new.recalculate(inplace=True)

    return composition_new


def moles_per_oxygen(moles):

    axis = [0, 1][isinstance(moles, pd.DataFrame)]

    O_number = e.oxygen_numbers(moles.elements)

    no_oxygen = O_number.index[O_number == 0]
    composition_new = _remove_elements(composition=moles, drop=no_oxygen)

    O_number = O_number.drop(no_oxygen)
    cations = e.cation_names(composition_new.elements)
    cation_number = e.cation_numbers(composition_new.elements)

    cations_normalised = cation_number.div(O_number)

    names_new = {
        ox: f"{cat}{Fraction(num).limit_denominator()}O"
        for cat, num, ox in zip(cations, cations_normalised, composition_new.elements)
    }
    names_new["total"] = "O_total"

    composition_new = (
        composition_new[composition_new.elements]
        .mul(O_number, axis=axis)
        .normalise(to=1)
    )

    return composition_new.rename(names_new, axis=axis)


def _anhydrous_composition(composition):

    composition_H2O = composition.copy()

    try:
        composition_H2O = composition_H2O.drop("H2O")
    except KeyError:
        composition_H2O = composition_H2O.drop(columns=["H2O"])

    return composition_H2O.recalculate().normalise()



from typing import Dict

import numpy as np


def birch_murnaghan_4th_order(
    V: float | np.ndarray, params: Dict = None
) -> float | np.ndarray:
    """
    Formulation from Katsura & Tange (2019)\ [26]_ (eq. 27). Needs parameters for:

    - pre-compression volume (V_0)
    - isothermal bulk modulus at standard temperature (K_0)
    - pressure derivative of the isothermal bulk modulus (Kprime_0)
    - second pressure derivative of the bulk modulus (Kprime_prime_0)


    Parameters
    ----------
    V : float, array-like
        volume
    params : dictionary, pandas Series
        fitted parameters 'V_0', 'K_0', 'Kprime_0', 'Kprime_prime_0'

    Returns
    -------
    pressure in same units as K_0
    """

    x = params["V_0"] / V

    part_1 = (3 / 2) * params["K_0"] * (np.power(x, 7 / 3) - np.power(x, 5 / 3))
    part_2 = 1 + (3 / 4) * (params["Kprime_0"] - 4) * (np.power(x, 2 / 3) - 1)
    part_3 = (1 / 24) * (
        9 * np.power(params["Kprime_0"], 2)
        - 63 * params["Kprime_0"]
        + 9 * params["K_0"] * params["Kprime_prime_0"]
        + 143
    )
    part_4 = np.power((np.power(x, 2 / 3) - 1), 2)

    return part_1 * (part_2 + part_3 * part_4)



"""
Sub-module with olivine-melt thermometers
"""

import inspect
import sys

import numpy as np
import pandas as pd

from MagmaPandas.magma_protocol import Magma
from MagmaPandas.parse_io import check_components
from MagmaPandas.tools.modify_compositions import _anhydrous_composition, _get_elements

errors = pd.Series({"putirka2007_4_dry": 45, "putirka2007_4_wet": 29})

components = {
    "putirka2007_4": ["MgO", "FeO", "Na2O", "K2O", "CaO", "SiO2", "K2O", "TiO2"]
}


def putirka2007_4(
    liquid: Magma,
    olivine: Magma,
    P_bar: float | pd.Series,
    offset: float = 0.0,
    **kwargs,
) -> pd.Series:
    """
    Olivine-liquid thermometer

    Equation 4 from Putirka (2007)\ [29]_ calculates liquidus temperatures based on olivine-melt pairs.
    Identical to equation 22 from Putirka (2008)\ [28]_.

    SEE = hydrous: 29, anhydrous: 45, total: 43

    Total Fe is expressed as FeO

    Parameters
    ----------
    liquid  : Magma
        melt compositions in oxide wt. %
    olivine : pandas Series, pandas Dataframe
        olivine compositions in oxide wt. %
    P_bar : float, pandas Series
        pressures in bar
    offset : float
        offset value in standard deviations. Temperatures are calculated as ``temnperature + offset * thermometer error (SEE)``.

    Returns
    -------
    temperatures : pd.Series
        olivine liquidus temperatures in Kelvin.
    """
    # from ..MagmaFrames import MagmaFrame
    # from ..MagmaSeries import MagmaSeries

    # if isinstance(composition, MagmaFrame):
    #     elements = composition.columns
    # elif isinstance(composition, MagmaSeries):
    #     elements = composition.index

    elements = _get_elements(liquid)

    composition = check_components(
        composition=liquid, components=components["putirka2007_4"]
    )

    # composition = liquid.copy(deep=True)
    # composition = composition.fillna(0.0)

    # oxides = set(["MgO", "FeO", "Na2O", "K2O", "CaO", "SiO2", "K2O", "TiO2"])
    # optional_oxides = {"MnO", "CoO", "NiO"}
    # absentOxides = oxides.difference(elements)
    # fill_oxides = optional_oxides.difference(elements)

    # if len(absentOxides) > 0:
    #     raise KeyError(f"{absentOxides} not found in melt")

    # All oxides as mole/cation fractions on an anhydrous basis
    # if len(fill_oxides) > 0:
    #     composition[list(fill_oxides)] = 0.0

    if "H2O" not in elements:
        error = errors["putirka2007_4_dry"]
        H2O = 0.0
    else:
        error = errors["putirka2007_4_wet"]
        H2O = composition["H2O"].copy()
        # moles are calculated on an anhydrous basis
        composition = _anhydrous_composition(composition)

    # Calculate molar oxide fractions
    liquid_moles = composition.moles()
    liquid_cations = composition.cations()
    olivine_cations = olivine.cations()

    C_NM = liquid_cations[["Fe", "Mn", "Mg", "Ca", "Co", "Ni"]].sum(
        axis=1
    )  # Fe = Fe total
    NF = 7 / 2 * np.log(1 - liquid_cations["Al"]) + 7 * np.log(1 - liquid_cations["Ti"])
    D_Mg = olivine_cations.loc[liquid_cations.index, "Mg"] / liquid_cations["Mg"]

    P_GPa = P_bar / 1e4

    numerator = 15294.6 + 1318.8 * P_GPa + 2.4834 * (P_GPa**2)
    denominator_a = 8.048 + 2.8532 * np.log(D_Mg) + 2.097 * np.log(1.5 * C_NM)
    denominator_b = (
        2.575 * np.log(3 * liquid_moles["SiO2"]) - 1.41 * NF + 0.222 * H2O + 0.5 * P_GPa
    )

    denominator = denominator_a + denominator_b

    T_K = (numerator / denominator) + 273.15

    T_K = T_K + error * offset

    return pd.Series(T_K, name="T_K").squeeze()


_module = sys.modules[__name__]
_funcs = inspect.getmembers(_module, inspect.isfunction)
# collect all ol-melt thermometers in a dictionary
olmelt_thermometers = {
    f[0]: f[1] for f in _funcs if f[1].__module__ == _module.__name__
}



import numpy as np

from MagmaPandas.parse_io.validate import _check_argument

"""
Re-written visual basic code from Holloway and Blank, 1994 (Reviews
in Mineralogy and Geochemistry, vol. 30).
Original python code from the supplementary material of:

C. M. Allison, K. Roggensack & A. B. Clarke (2022)
MafiCH: a general model for H2O–CO2 solubility in mafic magmas.
Contributions to Mineralogy and Petrology 177: 40
"""


class hollowayBlank:
    @staticmethod
    @_check_argument("species", [None, "H2O", "CO2"])
    def fugacity(T_K, P_bar, species):
        """
        Calculates CO2 or H2O fugacities at temperatures ``T_K`` and pressures ``P_bar``
        in bar
        """
        P0 = 4000

        # MRK below 4kb and Saxena above 4kb
        if P_bar > 4000 and species == "CO2":
            # Calculate FP at T and 4Kb
            iPUREG = hollowayBlank._RKCALC(T_K, P0, species)
            XLNF = hollowayBlank._Saxena(T_K, P_bar)
            PUREG = iPUREG + XLNF
        else:
            PUREG = hollowayBlank._RKCALC(T_K, P_bar, species)

        # Convert from LN fugacity to fugacity
        stdf = np.exp(PUREG)
        return stdf

    @staticmethod
    def _RKCALC(T_K, P_bar, species):
        """
        Calculation of pure gas MRK (modified Redlick-Kwong) properties following Holloway 1981, 1987
        P is in bar, T in Kelvin. Returns ln fugacity
        """

        R = 82.05736
        PBLN = np.log(P_bar)
        T_C = T_K - 273.15
        RXT = R * T_K
        RT = R * T_K**1.5 * 0.000001
        P_atmos = P_bar / 1.013  # 1e-6

        if species == "CO2":
            # Calculate T dependent MRK A parameter CO2
            ACO2M = 73.03 - 0.0714 * T_C + 2.157e-05 * T_C * T_C
            # Define MRK B parameter for CO2
            BSUM = 29.7
            ASUM = ACO2M / (BSUM * RT)

        elif species == "H2O":
            AH2OM = 115.98 - 0.0016295 * T_K - 1.4984e-05 * T_K**2
            BSUM = 14.5
            ASUM = AH2OM / (BSUM * RT)
        else:
            raise ValueError(
                f"species: {species} not recognised, please choose 'CO2' or 'H2O'"
            )

        BSUM = P_atmos * BSUM / RXT
        XLNFP = hollowayBlank._REDKW(BSUM, ASUM)
        # Convert to LN fugacity
        PUREG = XLNFP + PBLN
        return PUREG

    @staticmethod
    def _REDKW(BP, A2B):
        """
        The RK routine
        A routine to calc compressibility factor and fugacity coefficient with the
        Redlick-Kwong equation following Edmister (1968). This solution for
        supercritical fluid returns a value of 1 for FP if arguments are out of range.
        """

        TH = 1 / 3

        if A2B < 1e-10:
            A2B = 0.001

        RR = -A2B * BP * BP
        QQ = BP * (A2B - BP - 1)
        XN = QQ * TH + RR - 0.074074
        XM = QQ - TH
        XNN = XN * XN / 4
        XMM = XM * XM * XM / 27
        ARG = XNN + XMM

        if ARG > 0:
            X = np.sqrt(ARG)
            F = 1
            XN2 = -XN / 2
            iXMM = XN2 + X
            if iXMM < 0:
                F = -1
            XMM = F * ((F * iXMM) ** TH)
            F = 1
            iXNN = XN2 - X
            if iXNN < 0:
                F = -1
            XNN = F * ((F * iXNN) ** TH)
            Z = XMM + XNN + TH
            ZBP = Z - BP
            if ZBP < 0.000001:
                ZBP = 0.000001
            BPZ = 1 + BP / Z
            FP = Z - 1 - np.log(ZBP) - A2B * np.log(BPZ)
            if FP < -37 or FP > 37:
                FP = 0.000001

        elif ARG < 0:
            COSPHI = np.sqrt(-XNN / XMM)
            if XN > 0:
                COSPHI = -COSPHI
            TANPHI = np.sqrt(1 - COSPHI * COSPHI) / COSPHI
            PHI = np.arctan(TANPHI) * TH
            FAC = 2 * np.sqrt(-XM * TH)
            # sort for largest root
            R1 = np.cos(PHI)
            R2 = np.cos(PHI + 2.0944)
            R3 = np.cos(PHI + 4.18879)
            RH = R2
            if R1 > R2:
                RH = R1
            if R3 > RH:
                RH = R3
            Z = RH * FAC + TH
            ZBP = Z - BP
            if ZBP < 0.000001:
                ZBP = 0.000001
            BPZ = 1 + BP / Z
            FP = Z - 1 - np.log(ZBP) - A2B * np.log(BPZ)
            if FP < -37 or FP > 37:  # -37 > FP > 37
                FP = 0.000001

        else:
            FP = 1
            Z = 1
        XLNFP = FP
        return XLNFP

    @staticmethod
    def _Saxena(T_K, P_bar):
        """
        High pressure corresponding states routines from Saxena and Fei (1987) GCA
        Vol. 51, 783-791
        Returns natural np.log of the ratio F(P)/F(4000 bar) as XLNF array
        """

        # Define integration limit
        PO = 4000

        # Critical temperatures and pressures for CO2
        TR = T_K / 304.2
        PR = P_bar / 73.9
        PC = 73.9

        # Virial coefficients
        A = 2.0614 - 2.2351 / TR**2 - 0.39411 * np.log(TR)
        B = 0.055125 / TR + 0.039344 / TR**2
        C = -1.8935e-06 / TR - 1.1092e-05 / TR**2 - 2.1892e-05 / TR**3
        D = 5.0527e-11 / TR - 6.3033e-21 / TR**3

        # Calculate molar volume
        Z = A + B * PR + C * PR**2 + D * PR**3
        V = Z * 83.0117 * T_K / P_bar

        # integrate from PO (4000 bars) to P to calculate ln fugacity
        LNF = (
            A * np.log(P_bar / PO)
            + (B / PC) * (P_bar - PO)
            + (C / (2 * PC**2)) * (P_bar**2 - PO**2)
        )
        LNF = LNF + (D / (3 * PC**3)) * (P_bar**3 - PO**3)
        XLNF = LNF
        return XLNF



"""
Module with models for calculating melt |fO2| at various buffers.
"""

import numpy as np

from MagmaPandas.configuration import configuration
from MagmaPandas.fO2 import IW, QFM


def calculate_fO2(T_K, P_bar, **kwargs):

    fO2_buffer = kwargs.get("fO2_buffer", configuration.fO2buffer)
    dfO2 = kwargs.get("dfO2", configuration.dfO2)

    fO2_model = globals()[fO2_buffer]

    fO2 = fO2_model.calculate_fO2(logshift=dfO2, T_K=T_K, P_bar=P_bar)

    try:
        fO2 = fO2.astype(np.float32)
    except AttributeError:
        fO2 = np.float32(fO2)

    return fO2


def change_fO2_buffer(to: str, T_K: float | int, P_bar: float | int):

    current_fO2 = calculate_fO2(T_K=T_K, P_bar=P_bar)

    new_fO2_model = globals()[f"fO2_{to}"]

    new_fO2 = new_fO2_model(logshift=0, T_K=T_K, P_bar=P_bar)

    new_dfO2 = np.log10(current_fO2 / new_fO2)
    new_fO2 = np.float32(new_fO2 * 10**new_dfO2)

    if not np.isclose(new_fO2, current_fO2, rtol=0, atol=1e-5):
        raise ValueError("new fO2 does not match old fO2")

    configuration.fO2buffer = to
    configuration.dfO2 = round(new_dfO2, 2)

    print(configuration)



"""
Module with MagmaFrame classes and subclasses.
"""

from MagmaPandas.MagmaFrames.clinopyroxene import *
from MagmaPandas.MagmaFrames.ilmenite import *
from MagmaPandas.MagmaFrames.magmaFrame import *
from MagmaPandas.MagmaFrames.magnetite import *
from MagmaPandas.MagmaFrames.melt import *
from MagmaPandas.MagmaFrames.olivine import *
from MagmaPandas.MagmaFrames.plagioclase import *



"""
|CO2|-|H2O| solubility models from Iacono-Marziano et al. (2012)\ [33]_
"""

from typing import Tuple

import numpy as np
import pandas as pd
from scipy.optimize import root, root_scalar

from MagmaPandas.magma_protocol import Magma
from MagmaPandas.parse_io.validate import _check_argument, _check_setter
from MagmaPandas.volatile_solubility.solubility_baseclass import Solubility_model

"""
Equations from:

Iacono-Marziano G., Morizet Y., Le Trong E., Gaillard F. (2012) New experimental data and semi-empirical parameterization of |H2O|–|CO2| solubility in mafic melts. Geochimica et cosmochimica Acta. 97.
"""

parameter_options = ["hydrous_webapp", "hydrous_manuscript", "anhydrous"]
fugacity_options = ["ideal"]
activity_options = ["ideal"]
model_options = ["mixed", "h2o", "co2"]


class _meta_IaconoMarziano_configuration(type):
    def __init__(cls, *args, **kwargs):
        cls._parameters = "hydrous_webapp"
        cls._fugacity = "ideal"
        cls._activity = "ideal"
        cls._model = "mixed"

    @property
    def parameters(cls):
        return cls._parameters

    @parameters.setter
    @_check_setter(parameter_options)
    def parameters(cls, model: str):
        cls._parameters = model

    @property
    def fugacity(cls):
        return cls._fugacity

    @fugacity.setter
    @_check_setter(fugacity_options)
    def fugacity(cls, model: str):
        cls._fugacity = model

    @property
    def activity(cls):
        return cls._activity

    @activity.setter
    @_check_setter(activity_options)
    def activity(cls, model: str):
        cls._activity = model

    @property
    def model(cls):
        return cls._model

    @model.setter
    @_check_setter(model_options)
    def model(cls, model: str):
        cls._model = model

    @classmethod
    def reset(cls):
        cls._parameters = "hydrous_webapp"
        cls._fugacity = "ideal"
        cls._activity = "ideal"
        cls._model = "mixed"


class IaconoMarziano_configuration(metaclass=_meta_IaconoMarziano_configuration):
    @classmethod
    def reset(cls):
        cls._parameters = "hydrous_webapp"
        cls._fugacity = "ideal"
        cls._activity = "ideal"
        cls._model = "mixed"

    @classmethod
    def print(cls):
        """ """

        variables = {
            "Parameterisation": "_parameters",
            "Fugacity model": "_fugacity",
            "Activity model": "_activity",
            "Species model": "_model",
        }

        pad_left = 20
        pad_right = 20
        pad_total = pad_left + pad_right

        print(" Iacono-Marziano volatile solubility ".center(pad_total, "#"))
        print("".ljust(pad_total, "#"))
        print("Settings".ljust(pad_total, "_"))
        for param, model in variables.items():
            # model_attr = f"_IaconoMarziano_configuration{model}"
            print(f"{param:.<{pad_left}}{getattr(cls, model):.>{pad_right}}")
        print("\nCalibration range".ljust(pad_total, "_"))
        T_string = f"1373-1673\N{DEGREE SIGN}K"
        print(f"{'Temperature':.<{pad_left}}{T_string:.>{pad_right}}")
        print(f"{'Pressure':.<{pad_left}}{'0.1-5 kbar':.>{pad_right}}")
        print("\n")


H2O_coefficients = {
    "hydrous_webapp": {
        "a": 0.52096846,
        "b": 2.11575907,
        "B": -3.24443335,
        "C": -0.02238884,
    },
    "hydrous_manuscript": {
        "a": 0.53,
        "b": 2.35,
        "B": -3.37,
        "C": -0.02,
    },
    "anhydrous": {
        "a": 0.54,
        "b": 1.24,
        "B": -2.95,
        "C": 0.02,
    },
}

CO2_coefficients = {
    "hydrous": {
        "d_H2O": -16.4,
        "d_AI": 4.4,
        "d_FM": -17.1,
        "d_NK": 22.8,
        "a": 1.0,
        "b": 17.3,
        "B": -6.0,
        "C": 0.12,
    },
    "anhydrous": {
        "d_H2O": 2.3,
        "d_AI": 3.8,
        "d_FM": -16.3,
        "d_NK": 20.1,
        "a": 1.0,
        "b": 15.8,
        "B": -5.3,
        "C": 0.14,
    },
}


# Lower case names for classes, as to not mix up with variable names
class h2o(Solubility_model):
    @staticmethod
    def calculate_solubility(
        oxide_wtPercents: Magma,
        P_bar: float,
        T_K: float,
        x_fluid: float = 1.0,
        **kwargs,
    ) -> float:
        """
        Calculate melt |H2O| solubility according to equation 13.

        Parameters
        ----------
        oxide_wtPercents : MagmaSeries
            melt composition in oxide wt. %.
        P_bar : float
            Pressure in bar.
        T_K : float
            temperature in Kelvin
        x_fluid : float
            fraction of |H2O| in the fluid. Default value is 1.0

        Returns
        -------
        H2O : float
            melt |H2O| solubility in wt.%.
        """

        if not 1 >= x_fluid >= 0:
            raise ValueError(f"x_fluid: {x_fluid} is not between 0 and 1")
        if P_bar < 0:
            raise ValueError(f"Pressure is negative: '{P_bar}'")

        if any(i <= 0 for i in [P_bar, x_fluid]):
            return 0

        composition = oxide_wtPercents.copy()

        if IaconoMarziano_configuration.parameters == "anhydrous":
            # Solve equation 13
            return h2o._solubility(composition, x_fluid, P_bar, T_K)
        else:
            # Find equilibrium H2O content
            return np.float32(
                root_scalar(
                    h2o._solubility_rootFunction,
                    args=(composition, x_fluid, P_bar, T_K),
                    x0=1.0,
                    x1=2.0,
                ).root
            )

    @staticmethod
    def calculate_saturation(oxide_wtPercents: Magma, T_K: float, **kwargs) -> float:
        """
        Calculate melt |H2O| saturation pressure according to equation 13.

        Parameters
        ----------
        oxide_wtPercents : MagmaSeries
            melt composition in oxide wt. %. Needs to have an 'H2O' column.
        T_K : float
            temperature in Kelvin
        kwargs
            keyword arguments passed to :py:meth:`~MagmaPandas.volatile_solubility.models.IaconoMarziano.h2o.calculate_solubility`

        Returns
        -------
        P_saturation : float
            Saturation pressure in bar
        """

        if "H2O" not in oxide_wtPercents.index:
            raise ValueError("H2O not found in sample")
        if oxide_wtPercents["H2O"] < 0:
            raise ValueError(f"H2O lower than 0: {oxide_wtPercents['H2O']}")
        if oxide_wtPercents["H2O"] == 0.0:
            return 0.0

        composition = oxide_wtPercents.copy()

        # Upper limit of 15kbar
        upper_limit = 1.5e4
        # try:
        P_saturation = root_scalar(
            h2o._saturation_rootFunction,
            args=(composition, T_K, kwargs),
            bracket=[1e-15, upper_limit],
        ).root
        # except:
        #     P_saturation = np.nan
        return P_saturation

    @staticmethod
    def _solubility(oxide_wtPercents: Magma, x_fluid, P_bar, T_K):
        """
        Equation 13 from Iacono-Marziano et al. (2012)
        """

        coefficients = H2O_coefficients[IaconoMarziano_configuration.parameters]
        a, b, B, C = [coefficients[key] for key in ["a", "b", "B", "C"]]

        mol_fractions = oxide_wtPercents.moles()
        NBO_O = NBO_O_calculate(mol_fractions)

        if IaconoMarziano_configuration.fugacity == "ideal":
            P_H2O = x_fluid * P_bar

        H2O = np.exp(a * np.log(P_H2O) + b * NBO_O + B + C * P_bar / T_K)

        return H2O

    @staticmethod
    def _solubility_rootFunction(H2O, oxide_wtPercents: Magma, x_fluid, P_bar, T_K):
        """
        Compare input and output dissolved H2O
        """
        # Copy so pandas doesn't raise SettingWithCopyWarning
        composition = oxide_wtPercents.copy()
        composition["H2O"] = np.float32(H2O)

        return H2O - h2o._solubility(composition, x_fluid, P_bar, T_K)

    @staticmethod
    def _saturation_rootFunction(P_bar, oxide_wtPercents: Magma, T_K, kwargs):
        """ """
        composition = oxide_wtPercents.copy()
        #
        return composition["H2O"] - h2o.calculate_solubility(
            oxide_wtPercents=composition,
            P_bar=P_bar,
            T_K=T_K,
            **kwargs,
        )


# Lower case names for classes, as to not mix up with variable names
class co2(Solubility_model):
    @staticmethod
    def calculate_solubility(
        oxide_wtPercents: Magma,
        P_bar: float | np.ndarray,
        T_K: float | np.ndarray,
        x_fluid: float = 0.0,
        **kwargs,
    ) -> float:
        """
        Calculate melt |CO2| solubility according to equation 12.

        Parameters
        ----------
        oxide_wtPercents : MagmaSeries
            melt composition in oxide wt. %.
        P_bar : float, array-like
            Pressure in bar
        T_K : float, array-like
            temperature in Kelvin
        x_fluid : float
            fraction of |H2O| in the fluid. Default value is 0.0

        Returns
        -------
        solublity : float
            melt |CO2| solubility in wt. %.
        """

        if not 1 >= x_fluid >= 0:
            raise ValueError(f"x_fluid: {x_fluid} is not between 0 and 1")
        if P_bar < 0:
            raise ValueError(f"Pressure is negative: '{P_bar}'")

        if any(i <= 0 for i in [P_bar, (1 - x_fluid)]):
            return 0

        if "anhydrous" in IaconoMarziano_configuration.parameters:
            parameterisation = "anhydrous"
        else:
            parameterisation = "hydrous"

        coefficients = CO2_coefficients[parameterisation]
        d_H2O, d_AI, d_FM, d_NK, a, b, B, C = [
            coefficients[key]
            for key in ["d_H2O", "d_AI", "d_FM", "d_NK", "a", "b", "B", "C"]
        ]

        composition = oxide_wtPercents.copy()
        composition["H2O"] = h2o.calculate_solubility(composition, P_bar, T_K, x_fluid)
        mol_fractions = composition.moles()
        NBO_O = NBO_O_calculate(mol_fractions)

        if "Fe2O3" in mol_fractions.index:
            Fe2O3 = mol_fractions["Fe2O3"]
        else:
            Fe2O3 = 0.0

        if IaconoMarziano_configuration.fugacity == "ideal":
            P_CO2 = (1 - x_fluid) * P_bar

        x_AI = mol_fractions["Al2O3"] / (
            mol_fractions["CaO"] + mol_fractions["K2O"] + mol_fractions["Na2O"]
        )
        x_FM = mol_fractions["FeO"] + mol_fractions["MgO"] + 2 * Fe2O3
        x_NK = mol_fractions["Na2O"] + mol_fractions["K2O"]

        CO3_ppm = np.exp(
            mol_fractions["H2O"] * d_H2O
            + x_AI * d_AI
            + x_FM * d_FM
            + x_NK * d_NK
            + a * np.log(P_CO2)
            + b * NBO_O
            + B
            + C * P_bar / T_K
        )

        return CO3_ppm / 1e4

    @staticmethod
    def calculate_saturation(
        oxide_wtPercents: Magma,
        T_K: float,
        **kwargs,
    ) -> float:
        """
        Calculate melt |CO2| saturation pressure according to equation 12.

        Parameters
        ----------
        oxide_wtPercents : MagmaSeries
            melt composition in oxide wt. %. Needs to have a 'CO2' column.
        T_K : float
            temperature in Kelvin
        kwargs
            keyword arguments passed to :py:meth:`~MagmaPandas.volatile_solubility.models.IaconoMarziano.co2.calculate_solubility`

        Returns
        -------
        P_saturation : float
            Saturation pressure in bar
        """

        if "CO2" not in oxide_wtPercents.index:
            raise ValueError("CO2 not found in sample")
        if oxide_wtPercents["CO2"] < 0:
            raise ValueError(f"CO2 lower than 0: {oxide_wtPercents['CO2']}")
        if oxide_wtPercents["CO2"] == 0.0:
            return 0.0

        composition = oxide_wtPercents.copy()
        # Upper limit of 100kbar
        upper_limit = 1e5
        # try:
        P_saturation = root_scalar(
            co2._saturation_rootFunction,
            args=(composition, T_K, kwargs),
            bracket=[1e-10, upper_limit],
        ).root
        # except:
        #     P_saturation = np.nan
        return P_saturation

    @staticmethod
    def _saturation_rootFunction(P_bar, oxide_wtPercents: Magma, T_K, kwargs):
        """
        Compare calculated and sample CO2
        """
        composition = oxide_wtPercents.copy()
        return composition["CO2"] - co2.calculate_solubility(
            oxide_wtPercents=composition,
            P_bar=P_bar,
            T_K=T_K,
            **kwargs,
        )


class mixed(Solubility_model):
    @staticmethod
    @_check_argument("output", [None, "PXfl", "P", "Xfl"])
    def calculate_saturation(
        oxide_wtPercents: Magma, T_K: float, output: str = "P", **kwargs
    ) -> float | Tuple[float, float]:
        """
        Calculate volatile saturation pressure for systems with mixed |CO2|-|H2O| fluids.

        Parameters
        ----------
        oxide_wtPercents : MagmaSeries
            melt composision in oxide wt. %. Needs to have 'H2O' and 'CO2' columns.
        T_K : float
            Temperature in kelvin
        output : str
            Output format. 'P' for pressure only, 'Xfl' for |H2O| fluid fraction only and 'PXfl' for both.

        Returns
        -------
        saturation : float, (float, float)
            Depending on the value of ``output``: saturation pressure in bar, |H2O| fluid fraction or (saturation pressure, fluid fraction)
        """
        composition = oxide_wtPercents.copy()

        P_H2O_saturation = h2o.calculate_saturation(composition, T_K=T_K, x_fluid=1.0)
        P_CO2_saturation = co2.calculate_saturation(composition, T_K=T_K, x_fluid=0.0)

        if oxide_wtPercents["H2O"] <= 0.0:
            return P_CO2_saturation
        if oxide_wtPercents["CO2"] <= 0.0:
            return P_H2O_saturation

        P_guess = 0

        for species in (P_H2O_saturation, P_CO2_saturation):
            if np.isfinite(species):
                P_guess += species

        saturation = root(
            mixed._saturation_rootFunction,
            x0=[P_guess, 0.0],
            args=(composition, T_K),
        ).x

        if saturation[1] <= 0.0:
            saturation[0] = P_CO2_saturation
        elif saturation[1] >= 1.0:
            saturation[0] = P_H2O_saturation
        saturation[1] = np.clip(saturation[1], 0.0, 1.0)

        return_dict = {"P": saturation[0], "Xfl": saturation[1], "PXfl": saturation}
        return return_dict[output]

    @staticmethod
    @_check_argument("output", [None, "both", "CO2", "H2O"])
    def calculate_solubility(
        oxide_wtPercents: Magma,
        P_bar: float,
        T_K: float,
        x_fluid: float,
        output: str = "both",
        **kwargs,
    ) -> float | Tuple[float, float]:
        """
        Calculate volatile solubilities for systems with mixed |CO2|-|H2O| fluids.

        Parameters
        ----------
        oxide_wtPercents : MagmaSeries
            melt composision in oxide wt. %.
        P_bar : float
            pressure in bar
        T_K : float
            Temperature in kelvin
        x_fluid: float
            fraction of |H2O| in the fluid.
        output : str
            Output format. 'CO2' for |CO2| only, 'H2O' for |H2O| only and 'both' for both.

        Returns
        -------
        saturation : float, (float, float)
            Solubility in wt. %. Depending on the value of ``output``: |CO2|, |H2O| or (|CO2|, |H2O|).
        """
        if not 1 >= x_fluid >= 0:
            raise ValueError(f"x_fluid: {x_fluid} is not between 0 and 1")
        if P_bar < 0:
            raise ValueError(f"Pressure is negative: '{P_bar}'")

        composition = oxide_wtPercents.copy()
        H2O = h2o.calculate_solubility(composition, P_bar, T_K, x_fluid)
        composition["H2O"] = np.float32(H2O)
        CO2 = co2.calculate_solubility(composition, P_bar, T_K, x_fluid)

        return_dict = {"both": (H2O, CO2), "H2O": H2O, "CO2": CO2}
        return return_dict[output]

    @staticmethod
    def _saturation_rootFunction(P_x_fluid, oxide_wtPercents: Magma, T_K):
        """
        compare calculated with sample concentrations
        """

        P_bar, x_fluid = P_x_fluid
        # Keep x_fluid and P_bar within bounds
        x_fluid = np.clip(x_fluid, 0.0, 1.0)
        P_bar = np.clip(P_bar, a_min=1e-15, a_max=None)

        composition = oxide_wtPercents.copy()

        H2O = composition["H2O"]
        CO2 = composition["CO2"]

        sample_concentrations = np.array([H2O, CO2])
        calculated_concentrations = np.array(
            mixed.calculate_solubility(
                oxide_wtPercents=composition,
                P_bar=P_bar,
                T_K=T_K,
                x_fluid=x_fluid,
            )
        )

        return abs(calculated_concentrations - sample_concentrations)


def NBO_O_calculate(mol_fractions):
    """
    Non-bridging oxygen / oxygen ratio, following Marrochhi & Toplis (2005).
    See also Iacono-Marziano (2012) appendix.
    """
    if isinstance(mol_fractions, pd.Series):
        elements = mol_fractions.index
    elif isinstance(mol_fractions, pd.DataFrame):
        elements = mol_fractions.columns

    # All Fe is considered as FeO
    if "Fe2O3" in elements:
        Fe2O3 = mol_fractions["Fe2O3"]
    else:
        Fe2O3 = 0.0

    NBO = 2 * (
        mol_fractions["K2O"]
        + mol_fractions["Na2O"]
        + mol_fractions["CaO"]
        + mol_fractions["MgO"]
        + mol_fractions["FeO"]
        + 2 * Fe2O3
        - mol_fractions["Al2O3"]
    )
    O = (
        2 * mol_fractions["SiO2"]
        + 2 * mol_fractions["TiO2"]
        + 3 * mol_fractions["Al2O3"]
        + mol_fractions["MgO"]
        + mol_fractions["FeO"]
        + 2 * Fe2O3
        + mol_fractions["CaO"]
        + mol_fractions["Na2O"]
        + mol_fractions["K2O"]
    )

    if not IaconoMarziano_configuration.parameters == "anhydrous":
        NBO = NBO + 2 * mol_fractions["H2O"]
        O = O + mol_fractions["H2O"]

    return NBO / O



from importlib import resources
from typing import Dict

import matplotlib.patheffects as pe
import pandas as pd
from matplotlib.axes import Axes


def TAS(
    ax: Axes,
    labels=False,
    fontsize="medium",
    abbreviate=False,
    textkwargs: Dict = {},
    **kwargs,
):
    """Returns a line plot element of classification of volcanic rocks
    in total-alkali vs silica plots
    """

    # resources.open_text("geoplot.data", "TAS.csv")
    with resources.files("geoplot.data").joinpath("TAS.csv").open("r") as df:
        TAS = pd.read_csv(df)

    rock_labels = {
        "Picro-basalt": ["Picro\nbasalt", "Pi-Ba", [43, 1.5]],
        "Basalt": ["Basalt", "Ba", [49, 2.5]],
        "Basaltic andesite": ["Basaltic\nandesite", "Ba-An", [54.5, 2.5]],
        "Andesite": ["Andesite", "An", [60, 2.5]],
        "Dacite": ["Dacite", "Da", [68, 4]],
        "Trachy-basalt": ["Trachy-\nbasalt", "Tr-Ba", [49, 5.5]],
        "Basaltic trachy-andesite": [
            "Basaltic\ntrachy-\nandesite",
            "B\nTr-An",
            [52.5, 6.5],
        ],
        "Trachy-andesite": ["Trachy-\nandesite", "TrAn", [58.5, 8]],
        "Trachyte": ["Trachyte", "Tr", [64, 11]],
        "Tephrite": ["Tephrite", "Te", [45, 7]],
        "Phono-tephrite": ["Phono-\ntephrite", "Ph-Te", [49, 9.0]],
        "Tephri-phonolite": ["Tephri-\nphonolite", "Te-Ph", [53, 11]],
        "Phonolite": ["Phonolite", "Ph", [57, 15]],
        "Foidite": ["Foidite", "Fo", [45, 14]],
        "Rhyolite": ["Rhyolite", "Rh", [75, 8.5]],
    }

    if labels:
        for _, (fullname, abbreviation, coords) in rock_labels.items():
            label = [fullname, abbreviation][abbreviate]
            ax.text(
                *coords,
                label,
                fontsize=fontsize,
                fontfamily="monospace",
                clip_on=True,
                horizontalalignment="center",
                **textkwargs,
            )

    for id in TAS.id.unique():
        ax.plot(
            TAS.loc[TAS.id == id, "x"],
            TAS.loc[TAS.id == id, "y"],
            "-",
            color="k",
            **kwargs,
        )



import pandas as pd
from typing_extensions import Self

from MagmaPandas.configuration import configuration
from MagmaPandas.enums import Unit
from MagmaPandas.Fe_redox.Fe3Fe2_models import Fe3Fe2_models_dict
from MagmaPandas.MagmaFrames.magmaFrame import MagmaFrame
from MagmaPandas.MagmaFrames.melt import Melt
from MagmaPandas.MagmaSeries import MagmaSeries


class Olivine(MagmaFrame):
    """
    Subclass of :py:class:`~MagmaPandas.MagmaFrames.magmaFrame.MagmaFrame` extended with olivine specific methods.
    """

    # @property
    # def _constructor(self):
    #     """This is the key to letting Pandas know how to keep
    #     derivatives of `MagmaBase` the same type as yours.  It should
    #     be enough to return the name of the Class.  However, in
    #     some cases, `__finalize__` is not called and `new attributes` are
    #     not carried over.  We can fix that by constructing a callable
    #     that makes sure to call `__finalize__` every time."""

    #     def _c(*args, weights=None, **kwargs):
    #         if weights is None:
    #             weights = self._weights.copy(deep=True)
    #         return Olivine(*args, weights=weights, **kwargs).__finalize__(self)

    #     return _c

    @property
    def forsterite(self) -> pd.Series:
        """
        Forsterite contents
        """
        if self._units == Unit.WT_PERCENT:
            moles = self.moles()
        else:
            moles = self

        type = self._datatype.value

        Mg = {"oxide": "MgO", "cation": "Mg"}
        Fe = {"oxide": "FeO", "cation": "Fe"}
        self.recalculate(inplace=True)
        return pd.Series(
            moles[Mg[type]] / (moles[Fe[type]] + moles[Mg[type]]),
            name="Fo#",
        )

    @property
    def formula(self) -> Self:
        """
        Mineral formulas normalised to 4 O p.f.u.
        """
        return self.mineral_formula(O=4)

    def calculate_FeMg_Kd(
        self, melt_wtpc: Melt | MagmaSeries, T_K, P_bar=1, **kwargs
    ) -> pd.Series:
        """
        Calculate Fe-Mg exchange coefficients (Kd) based on measured olivine and melt compositions as (Fe2+/Mg)olivine / (Fe2+/Mg)liquid
        """

        if (not self.index.equals(melt_wtpc.index)) & (
            melt_wtpc.isinstance(pd.DataFrame)
        ):
            raise AttributeError("olivine and melt indeces do not match")

        dfO2 = kwargs.get("dfO2", configuration.dfO2)
        Fe3Fe2_model_name = kwargs.get("Fe3Fe2_model", configuration.Fe3Fe2_model)
        Fe3Fe2_model = Fe3Fe2_models_dict[Fe3Fe2_model_name]

        melt_mol_fractions = melt_wtpc.moles()
        olivine_mol_fractions = self.moles()

        Fe3Fe2 = Fe3Fe2_model.calculate_Fe3Fe2(
            mol_fractions=melt_mol_fractions,
            T_K=T_K,
            P_bar=P_bar,
            dfO2=dfO2,
            **kwargs,
        )

        Fe_melt_total = 1 + Fe3Fe2
        Fe2_FeTotal_melt = 1 / Fe_melt_total
        Fe2_melt = melt_mol_fractions["Fe"] * Fe2_FeTotal_melt

        Kd = (olivine_mol_fractions["Fe"] / olivine_mol_fractions["Mg"]) / (
            Fe2_melt / melt_mol_fractions["Mg"]
        )

        return Kd



"""
Global configuration of MagmaPandas settings.
"""

from MagmaPandas.Fe_redox.Fe3Fe2_models import Fe3Fe2_models_dict
from MagmaPandas.Kd.Ol_melt.FeMg.Kd_models import Kd_olmelt_FeMg_models_dict
from MagmaPandas.parse_io.validate import _check_setter
from MagmaPandas.thermometers.melt import melt_thermometers_dict
from MagmaPandas.volatile_solubility.volatile_solubility_models import (
    volatile_solubility_models_dict,
)

Fe3Fe2_models = list(Fe3Fe2_models_dict.keys())
Kd_ol_FeMg_models = list(Kd_olmelt_FeMg_models_dict.keys())
fO2_buffers = ["QFM", "IW"]
melt_thermometers = list(melt_thermometers_dict.keys())
volatile_solubility_models = list(volatile_solubility_models_dict.keys())
volatile_species_options = ["co2", "h2o", "mixed"]

_variables = {
    "fO2 buffers": fO2_buffers,
    "Melt Fe3+/Fe2+ models": Fe3Fe2_models,
    "Ol-melt Fe-Mg Kd models": Kd_ol_FeMg_models,
    "Melt thermometers": melt_thermometers,
    "Volatile solubility models": volatile_solubility_models,
    "Volatile species": volatile_species_options,
}

# _variables = {
#     "fO2 buffers": fO2_buffers,
#     "Melt Fe3+/Fe2+ models": list(Fe3Fe2_models.keys()),
#     "Ol-melt Fe-Mg Kd models": list(Kd_olmelt_FeMg_models.keys),  # Kd_ol_FeMg_models,
#     "Melt thermometers": list(thermometers_melt.keys()),  # melt_thermometers,
#     "Volatile solubility models": list(
#         volatile_solubility_models.keys()
#     ),  # volatile_solubility_models,
#     "Volatile species": volatile_species_options,
# }

_names_length = max([len(i) for i in _variables.keys()]) + 5
_pad_right = max([len(", ".join(i)) for i in _variables.values()])
# _pad_total = _names_length + _pad_right

[len(", ".join(i)) for i in _variables.values()]
_pad_total = 70
_pad_right = _pad_total - _names_length
_new_line = "\n"

configuration_options = (
    f"{_new_line}{' MagmaPandas ':#^{_pad_total}}"
    f"{_new_line}{'':#^{_pad_total}}"
    f"{_new_line}{'Configuration options':_<{_pad_total}}"
)

for param, values in _variables.items():

    val_string_2 = ""
    idxs = [0, 0]
    start_idx = 0
    while idxs[-1] < len(values):

        for i, _ in enumerate(values):
            if len(", ".join(values[start_idx : i + 1])) > (_pad_total - _names_length):
                idxs.insert(-1, i)
                start_idx = i
                continue
            idxs[-1] = i + 1

    if len(idxs) > 2:
        val_string_1 = ", ".join(values[: idxs[1]])
        val_string_2 = f""
        for i, idx in enumerate(idxs[2:]):
            val_string_2 += (
                f"{_new_line}{', '.join(values[idxs[i+1]:idx]):>{_pad_total}}"
            )
        val_string_2 += f"{_new_line}"
    else:
        val_string_1 = ", ".join(values)

    line = f"{_new_line}{param:.<{_names_length}}{val_string_1:.>{_pad_right}}{val_string_2}"
    configuration_options += line

configuration_options += f"{_new_line}{'':#^{_pad_total}}"


class _meta_configuration(type):
    def __init__(cls, *args, **kwargs):
        cls._fO2buffer = "QFM"
        cls._dfO2 = 1
        cls._Kd_model = "toplis"
        cls._Fe3Fe2_model = "Sun2024"
        cls._melt_thermometer = "putirka2008_15"
        cls._volatile_solubility = "IaconoMarziano"
        cls._volatile_species = "mixed"

    @property
    def fO2buffer(cls):
        return cls._fO2buffer

    @fO2buffer.setter
    @_check_setter(fO2_buffers)
    def fO2buffer(cls, value: str):
        cls._fO2buffer = value

    @property
    def dfO2(cls):
        return cls._dfO2

    @dfO2.setter
    def dfO2(cls, value: float | int):
        cls._dfO2 = value

    @property
    def Kd_model(cls):
        return cls._Kd_model

    @Kd_model.setter
    @_check_setter(Kd_ol_FeMg_models)
    def Kd_model(cls, model: str):
        cls._Kd_model = model

    @property
    def Fe3Fe2_model(cls):
        return cls._Fe3Fe2_model

    @Fe3Fe2_model.setter
    @_check_setter(Fe3Fe2_models)
    def Fe3Fe2_model(cls, model: str):
        cls._Fe3Fe2_model = model

    @property
    def melt_thermometer(cls):
        return cls._melt_thermometer

    @melt_thermometer.setter
    @_check_setter(melt_thermometers)
    def melt_thermometer(cls, model: str):
        cls._melt_thermometer = model

    @property
    def volatile_solubility(cls):
        return cls._volatile_solubility

    @volatile_solubility.setter
    @_check_setter(volatile_solubility_models)
    def volatile_solubility(cls, model: str):
        cls._volatile_solubility = model

    @property
    def volatile_species(cls):
        return cls._volatile_species

    @volatile_species.setter
    @_check_setter(volatile_species_options)
    def volatile_species(cls, model: str):
        cls._volatile_species = model

    def __str__(cls):

        variables = {
            "fO2 buffer": "_fO2buffer",
            "\u0394fO2": "_dfO2",
            "Melt Fe3+/Fe2+": "_Fe3Fe2_model",
            "Kd Fe-Mg ol-melt": "_Kd_model",
            "Melt thermometer": "_melt_thermometer",
            "Volatile solubility model": "_volatile_solubility",
            "Volatile species": "_volatile_species",
        }

        names_length = max([len(i) for i in variables.keys()]) + 5
        pad_right = 15
        pad_total = names_length + pad_right
        new_line = "\n"

        message = (
            f"{new_line}{' MagmaPandas ':#^{pad_total}}"
            f"{new_line}{'':#^{pad_total}}"
            f"{new_line}{'General settings':_<{pad_total}}"
            # f"{new_line}{'fO2 buffer':.<{names_length}}{'QFM':.>{pad_right}}"
        )

        parameter_settings = ""
        for param, value in variables.items():
            parameter_settings += f"{new_line}{param:.<{names_length}}{str(getattr(cls, value)):.>{pad_right}}"

        return message + parameter_settings + f"{new_line}{'':#^{pad_total}}{new_line}"


class configuration(metaclass=_meta_configuration):
    """
    Class for configuring global settings in MagmaPandas.

    Attributes
    ----------
    dQFM    : int, float
        Log units shift of the QFM |fO2| buffer. Default value: 1
    Kd_model : str
        Olivine-melt Fe-Mg partitioning model.
    Fe3Fe2_model : str
        Melt |Fe3Fe2| model.
    melt_thermometer : str
        Melt-only thermometer.
    volatily_solubility : str
        |CO2|-|H2O| solubility model.
    volatile_species : str
        Fluid phase species. Options: 'h2o', 'co2' or 'mixed'. Default value: 'mixed'
    """

    @classmethod
    def available_models(cls):
        """
        Show available models for all parameters
        """
        print(configuration_options)

    @classmethod
    def reset(cls):
        """
        Reset to default values
        """
        cls._fO2buffer = "QFM"
        cls._dfO2 = 1
        cls._Kd_model = "toplis"
        cls._Fe3Fe2_model = "Sun2024"
        cls._melt_thermometer = "putirka2008_15"
        cls._volatile_solubility = "IaconoMarziano"
        cls._volatile_species = "mixed"



import inspect
import sys
from typing import Optional

import numpy as np
import pandas as pd
from scipy.constants import Avogadro, R
from scipy.optimize import fsolve

from MagmaPandas.EOSs.birch_murnaghan import birch_murnaghan_4th_order
from MagmaPandas.Fe_redox.Fe3Fe2_baseclass import Fe3Fe2_model
from MagmaPandas.Fe_redox.Fe3Fe2_errors import (
    error_params_1bar,
    error_params_high_pressure,
)
from MagmaPandas.parse_io import check_components, make_equal_length


def _is_Fe3Fe2_model(cls):
    """
    check if class inherits from Kd_model
    """
    try:
        return isinstance(cls(), Fe3Fe2_model)
    except TypeError:
        return False


class borisov(Fe3Fe2_model):
    """
    Calculate melt |Fe3Fe2| ratios according to equation 4 from Borisov et al. (2018)\ [1]_.
    """

    components = ["SiO2", "TiO2", "MgO", "CaO", "Na2O", "K2O", "Al2O3", "P2O5"]

    error_params = [0.03152663, 0.02589086, 0.6874305, 7.73489745]

    @classmethod
    def calculate_Fe3Fe2(
        cls, melt_mol_fractions: pd.DataFrame, T_K, fO2, *args, **kwargs
    ):
        """
        Calculate melt |Fe3Fe2| ratios.

        Parameters
        ----------
        mol_fractions   :   :py:class:`Pandas DataFrame <pandas:pandas.DataFrame>`
            Melt composition in oxide mol fractions
        T_K :   float, array-like
            temperature in Kelvin
        fO2 :   float, array-like
            Oxygen fugacity

        Returns
        -------
        float, array-like
            melt |Fe3Fe2| ratio
        """

        moles = check_components(
            composition=melt_mol_fractions, components=cls.components
        )

        part1 = (
            0.207 * np.log10(fO2)
            + 4633.3 / T_K
            - 0.445 * moles["SiO2"]
            - 0.900 * moles["TiO2"]
            + 1.532 * moles["MgO"]
        )
        part2 = (
            0.314 * moles["CaO"]
            + 2.030 * moles["Na2O"]
            + 3.355 * moles["K2O"]
            - 4.851 * moles["P2O5"]
        )
        part3 = (
            -3.081 * moles["SiO2"] * moles["Al2O3"]
            - 4.370 * moles["SiO2"] * moles["MgO"]
            - 1.852
        )

        return 10 ** (part1 + part2 + part3)

    @classmethod
    def get_error(cls, Fe3Fe2, pressure: Optional[pd.Series] = None, *args, **kwargs):

        return super().get_error(
            Fe3Fe2=Fe3Fe2,
            error_params_1bar=error_params_1bar,
            error_params_high_pressure=error_params_high_pressure,
            pressure=pressure,
        )


class kressCarmichael(Fe3Fe2_model):
    """
    Calculate melt |Fe3Fe2| ratios according to equation 7 from Kress and Carmichael (1991)\ [2]_.
    """

    components = ["Al2O3", "FeO", "CaO", "Na2O", "K2O"]

    # Parameters from table 7
    a = 0.196
    b = 1.1492e4
    c = -6.675

    dCoefficients = pd.Series(
        {"Al2O3": -2.243, "FeO": -1.828, "CaO": 3.201, "Na2O": 5.854, "K2O": 6.215},
        name="dCoeff",
    )

    e = -3.36
    f = -7.01e-7
    g = -1.54e-10
    h = 3.85e-17
    T0 = 1673

    FeO_error = 0.21
    Fe2O3_error = 0.42

    error_params = [7.34976440e-02, 1.81801171e-02, 9.87644316e-01, 2.27525692e02]

    @classmethod
    def calculate_Fe3Fe2(cls, melt_mol_fractions, T_K, fO2, P_bar):
        """
        Calculate melt |Fe3Fe2| ratios\.

        Parameters
        ----------
        mol_fractions   :  :py:class:`Pandas DataFrame <pandas:pandas.DataFrame>`
            melt composition in oxide mol fractions
        T_K        :     float, array-like
            temperature in Kelvin
        fO2     :      float, array-like
            Oxygen fugacity
        P_bar      :      float, array-like
            Pressure in bar


        Returns
        -------
        float, array-like
            melt |Fe3Fe2| ratio
        """

        moles = check_components(
            composition=melt_mol_fractions, components=cls.components
        )

        P_Pa = P_bar / 1e5

        axis = [0, 1][isinstance(moles, pd.DataFrame)]

        sumComponents = moles[cls.components].mul(cls.dCoefficients).sum(axis=axis)

        part1 = cls.a * np.log(fO2) + cls.b / T_K + cls.c + sumComponents
        part2 = cls.e * (1 - cls.T0 / T_K - np.log(T_K / cls.T0))
        part3 = (
            cls.f * P_Pa / T_K
            + cls.g * ((T_K - cls.T0) * P_Pa) / T_K
            + cls.h * P_Pa**2 / T_K
        )

        return 2 * np.exp(part1 + part2 + part3)

    @classmethod
    def get_error(cls, Fe3Fe2, pressure: Optional[pd.Series] = None, *args, **kwargs):

        return super().get_error(
            Fe3Fe2=Fe3Fe2,
            error_params_1bar=error_params_1bar,
            error_params_high_pressure=error_params_high_pressure,
            pressure=pressure,
        )


class jayasuriya(Fe3Fe2_model):
    """
    Jayasuriya et al. (2004)\ [3]_ equation 12.
    """

    components = ["MgO", "CaO", "Na2O", "K2O", "Al2O3", "P2O5", "FeO"]

    # Parameters from table 7
    a = 0.1967
    b = 12420
    c = 7.054

    dCoefficients = pd.Series(
        {
            "MgO": -0.487,
            "CaO": 2.201,
            "Na2O": 6.610,
            "K2O": 8.214,
            "Al2O3": -3.781,
            "P2O5": -62.79,
            "FeO": 1.377,
        },
        name="dCoeff",
    )

    error_params = [2.10272743e-01, -1.27359105e-02, 9.82853119e-01, 1.75114023e02]

    @classmethod
    def calculate_Fe3Fe2(
        cls, melt_mol_fractions, T_K, fO2, *args, **kwargs
    ) -> float | np.ndarray:

        moles = check_components(
            composition=melt_mol_fractions, components=cls.components
        )

        axis = [0, 1][isinstance(moles, pd.DataFrame)]

        sumComponents = moles[cls.components].mul(cls.dCoefficients).sum(axis=axis)

        return 2 * np.exp(cls.a * np.log(fO2) + 12420 / T_K - cls.c + sumComponents)

    @classmethod
    def get_error(cls, Fe3Fe2, pressure: Optional[pd.Series] = None, *args, **kwargs):

        return super().get_error(
            Fe3Fe2=Fe3Fe2,
            error_params_1bar=error_params_1bar,
            error_params_high_pressure=error_params_high_pressure,
            pressure=pressure,
        )


class putirka2016_6b(Fe3Fe2_model):
    """
    Putirka (2016)\ [4]_ equation 6b.
    """

    components = ["Na2O", "K2O", "Al2O3", "SiO2", "CaO"]

    a = 6.53
    b = 10813.8
    c = 0.19
    d = 12.4
    e = 3.44
    f = 4.15

    error_params = [1.87571331e-01, 2.58857703e-03, 8.65663335e-01, 2.43304009e01]

    @classmethod
    def calculate_Fe3Fe2(
        cls, melt_mol_fractions, T_K, fO2, *args, **kwargs
    ) -> float | np.ndarray:

        moles = check_components(
            composition=melt_mol_fractions, components=cls.components
        )

        axis = [0, 1][isinstance(moles, pd.DataFrame)]

        part1 = -cls.a + cls.b / T_K
        part2 = cls.c * np.log(fO2) + 12.4 * moles[["Na2O", "K2O"]].sum(axis=axis)
        part3 = (
            -cls.e * (moles["Al2O3"] / moles[["Al2O3", "SiO2"]].sum(axis=axis))
            + cls.f * moles["CaO"]
        )

        return 2 * np.exp(part1 + part2 + part3)

    @classmethod
    def get_error(cls, Fe3Fe2, pressure: Optional[pd.Series] = None, *args, **kwargs):

        return super().get_error(
            Fe3Fe2=Fe3Fe2,
            error_params_1bar=error_params_1bar,
            error_params_high_pressure=error_params_high_pressure,
            pressure=pressure,
        )


class putirka2016_6c(Fe3Fe2_model):
    """
    Putirka (2016)\ [4]_ equation 6c.
    """

    components = [
        "Al2O3",
        "Na2O",
        "K2O",
        "CaO",
        "MgO",
        "SiO2",
        "TiO2",
        "Cr2O3",
        "FeO",
        "MnO",
        "P2O5",
    ]

    a = -6.75
    b = 10634.9
    c = 0.195
    d = 7.9
    e = -4.6
    f = 0.54
    g = -53.4
    h = 1.07

    error_params = [-3.12797578e-02, 6.71834626e-02, 9.81552910e-01, 1.30748389e02]

    @classmethod
    def calculate_Fe3Fe2(
        cls, melt_mol_fractions, T_K, fO2, *args, **kwargs
    ) -> float | np.ndarray:

        moles = check_components(
            composition=melt_mol_fractions, components=cls.components
        )
        axis = [0, 1][isinstance(moles, pd.DataFrame)]

        NBO_T = cls._NBO_T(cations=moles.cations())

        part_1 = cls.a + cls.b / T_K + cls.c * np.log(fO2)
        part_2 = cls.d * moles[["Na2O", "K2O"]].sum(axis=axis) + cls.e * moles["MgO"]
        part_3 = (
            cls.f * (moles["MgO"] / moles[["MgO", "FeO"]].sum(axis=axis))
            + cls.g * moles["P2O5"]
            + cls.h * NBO_T
        )

        return 2 * np.exp(part_1 + part_2 + part_3)

    @staticmethod
    def _NBO_T(cations):

        axis = [0, 1][isinstance(cations, pd.DataFrame)]

        if axis == 1:
            Al_IV = cations.apply(
                lambda row: min(
                    row["Al"],
                    row[["Na", "K"]].sum() + 2 * row[["Ca", "Mg"]].sum(),
                ),
                axis=1,
            )
        else:
            Al_IV = min(
                cations["Al"],
                cations[["Na", "K"]].sum() + 2 * cations[["Ca", "Mg"]].sum(),
            )

        tetrahedral = cations[["Si", "Ti"]].sum(axis=axis) + Al_IV
        O = (
            2 * cations[["Si", "Ti"]].sum(axis=axis)
            + 1.5 * cations[["Al", "Cr"]].sum(axis=axis)
            + cations[["Fe", "Mn", "Mg", "Ca"]].sum(axis=axis)
            + 0.5 * cations[["Na", "K"]].sum(axis=axis)
            + 2.5 * cations["P"]
        )
        NBO = 2 * O - 4 * tetrahedral
        return NBO / tetrahedral

    @classmethod
    def get_error(cls, Fe3Fe2, pressure: Optional[pd.Series] = None, *args, **kwargs):

        return super().get_error(
            Fe3Fe2=Fe3Fe2,
            error_params_1bar=error_params_1bar,
            error_params_high_pressure=error_params_high_pressure,
            pressure=pressure,
        )


class Deng2020(Fe3Fe2_model):
    """
    Deng et al. (2020)\ [5]_
    """

    components = [
        "SiO2",
        "Al2O3",
        "FeO",
        "MgO",
        "CaO",
        "K2O",
        "Na2O",
        "TiO2",
        "P2O5",
    ]
    gibbs_parameters = pd.Series(
        {
            "a": -3.310e5,
            "b": -190.379,
            "c": 14.785,
            "d": -1.649e-3,
            "e": 9.348e6,
            "f": 1.077e4,
        }
    )

    margules = pd.Series(  # fit 3
        {
            "Mg": 68629,
            "Si": 4601,
            "Al": 40923,
            "Ca": -58109,
            "Na": 0,
            "K": -59584,
            "P": 0,
            "Ti": 0,
        }
    )
    Fe_margules = -14210  # fit 3

    eos_params = pd.DataFrame(
        [
            [1180.114014, 1204.763652, 1192.011066, 1256.727179],
            [26.94713861, 23.19530062, 23.95435759, 16.12613905],
            [2.802531871, 3.216089358, 3.32104996, 4.584011905],
            [0.012313472, 0.009340183, -0.008912497, -0.177152954],
        ],
        columns=list(zip(*[["12.5molpc"] * 2 + ["25molpc"] * 2, ["Fe2", "Fe3"] * 2])),
        index=["V_0", "K_0", "Kprime_0", "Kprime_prime_0"],
    )

    thermal_pressure_params = pd.DataFrame(
        [
            [35.79397483, 34.52616394, 31.34712676, 30.38414264],
            [71.10313668, 68.64429623, 62.48520005, 59.10950152],
            [36.59545225, 35.27069116, 32.4675829, 29.64971394],
        ],
        columns=list(zip(*[["12.5molpc"] * 2 + ["25molpc"] * 2, ["Fe2", "Fe3"] * 2])),
        index=["a", "b", "c"],
    )
    formula_units = {"12.5molpc": 2.0, "25molpc": 4.0}
    T_K_ref = 3000
    A3_to_cm3 = 1e-24

    error_params = [2.17554822e-01, -4.67257455e-03, 9.83474292e-01, 2.27563235e02]

    @classmethod
    def calculate_Fe3Fe2(
        cls,
        melt_mol_fractions,
        T_K: float | np.ndarray,
        P_bar: float | np.ndarray,
        fO2: float | np.ndarray,
        melt_Fe: str = "12.5molpc",
        Fe3Fe2_init=0.3,
        total_Fe="FeO",
        **kwargs,
    ):

        moles = (
            melt_mol_fractions.to_frame().T
            if isinstance(melt_mol_fractions, pd.Series)
            else melt_mol_fractions.copy()
        )  # force to DataFrame

        # TODO  CHECK UNITS: BARS, PASCAL OR GPA??

        dVdP = cls._dVdP(T_K=T_K, P_bar=P_bar, melt_Fe=melt_Fe, **kwargs)
        gibbs0 = cls._gibbs0(T_K)

        try:
            if len(moles) != (l := len(T_K)):
                moles = moles.loc[moles.index.repeat(l)].reset_index(drop=True)
        except TypeError:
            pass
        # force everything to an iteratble
        T_K, fO2, gibbs0, dVdP = make_equal_length(T_K, fO2, gibbs0, dVdP)

        Fe3Fe2_func = (
            lambda Fe3Fe2_guess, moles, T, fO2, G, dVdP: cls._Fe3Fe2_solver(
                melt_mol_fractions=moles,
                T_K=T,
                fO2=fO2,
                gibbs0=G,
                dVdP=dVdP,
                Fe3Fe2=Fe3Fe2_guess[0],
                total_Fe=total_Fe,
            )
            - Fe3Fe2_guess
        )

        Fe3Fe2 = pd.Series(dtype=float)

        # for (n, X), T, fO2, G, VP in zip(moles.iterrows(), T_K, fO2_GPa, gibbs0, dVdP):
        #     Fe3Fe2.loc[n] = cls._Fe3Fe2_solver(
        #         melt_mol_fractions=X, T_K=T, fO2_GPa=fO2, gibbs0=G, dVdP=VP
        #     )

        for (n, X), T, f, G, VP in zip(moles.iterrows(), T_K, fO2, gibbs0, dVdP):
            Fe3Fe2.loc[n] = fsolve(Fe3Fe2_func, x0=Fe3Fe2_init, args=(X, T, f, G, VP))[
                0
            ]

        return Fe3Fe2.squeeze()

    @classmethod
    def _Fe3Fe2_solver(
        cls,
        melt_mol_fractions: pd.Series,
        T_K: float,
        fO2: float,
        gibbs0: float,
        dVdP: float,
        Fe3Fe2=0.3,
        total_Fe="FeO",
    ) -> float | np.ndarray:
        """
        Equation 3 solved for Fe3Fe2
        """

        moles = check_components(
            composition=melt_mol_fractions, components=cls.components
        )
        cations = moles.FeO_Fe2O3_calc(Fe3Fe2, wtpc=False, total_Fe=total_Fe).cations()
        Fe_activities = cls._Fe_activities(cations, T_K)

        Fe3Fe2 = np.exp(-(gibbs0 + dVdP) / (R * T_K) - Fe_activities + np.log(fO2) / 4)

        return Fe3Fe2

    @classmethod
    def _gibbs0(cls, T_K):
        """
        Calculates the Gibbs free energy of equation 1 at reference P and T. Supplementary Note 1
        """

        return (
            cls.gibbs_parameters["a"]
            + cls.gibbs_parameters["b"] * T_K
            + cls.gibbs_parameters["c"] * T_K * np.log(T_K)
            + cls.gibbs_parameters["d"] * (T_K**2)
            + cls.gibbs_parameters["e"] / T_K
            + cls.gibbs_parameters["f"] * np.sqrt(T_K)
        )

    @classmethod
    def _thermal_pressure_coeff(cls, V, V0, params):
        """
        Calculates the thermal pressure coefficient in equation 2.
        """

        return (
            params["a"] - params["b"] * (V / V0) + params["c"] * np.power(V / V0, 2)
        ) / 1000

    @classmethod
    @np.vectorize(excluded=["cls", "phase", "melt_Fe"])
    def _calculate_volume(
        cls, T_K: float, P_bar: float, phase: str, melt_Fe: str
    ) -> float:
        """
        Equation 2 solved for volume.

        Returns
        -------
        volume  :   float
            volume in Angstrom^3
        """
        P_GPa = P_bar / 1e4
        eos_params = cls.eos_params[(f"{melt_Fe}", f"{phase}")]
        thermal_pressure_params = cls.thermal_pressure_params[
            (f"{melt_Fe}", f"{phase}")
        ]

        v_init = eos_params["V_0"] - 6 * P_GPa

        func = (
            lambda v: (
                birch_murnaghan_4th_order(V=v, params=eos_params)
                + cls._thermal_pressure_coeff(
                    V=v, V0=eos_params["V_0"], params=thermal_pressure_params
                )
                * (T_K - cls.T_K_ref)
            )
            - P_GPa
        )

        V_sol = fsolve(func, x0=v_init)[0]

        V_sol / cls.formula_units[
            melt_Fe
        ]  # do not delete this line; for some reason that results in divide by 0 warnings from numpy. Don't ask me why :(

        return V_sol

    @classmethod
    def _deltaV(cls, T_K, P_bar, melt_Fe):
        """
        Volume change across equation 1.

        Returns
        -------
        dV : float
            delta Volume in cm^3/mol (Fe3 - Fe2).
        """
        if isinstance(T_K, (float, int)):
            T_K = np.ones_like(P_bar) * T_K

        Fe2_volume = (
            cls._calculate_volume(T_K, P_bar, "Fe2", melt_Fe)
            / cls.formula_units[melt_Fe]
            * Avogadro
            * cls.A3_to_cm3
        )  # cm3/mol

        Fe3_volume = (
            cls._calculate_volume(T_K, P_bar, "Fe3", melt_Fe)
            / cls.formula_units[melt_Fe]
            * Avogadro
            * cls.A3_to_cm3
        )  # cm3/mol

        return Fe3_volume - Fe2_volume

    @classmethod
    @np.vectorize(excluded=["cls", "melt_Fe", "P_bar_min", "P_bar_step"])
    def _dVdP(
        cls,
        T_K: float | np.ndarray,
        P_bar: float | np.ndarray,
        melt_Fe: str,
        Pbar_min=1.0,
        Pbar_step=5e2,
        **kwargs,
    ):
        """
        delta V integrated for pressure

        Returns
        -------
        dVdP : float, array-like
            dVdP in m3.Pascal
        """

        P_array = np.arange(Pbar_min, P_bar + Pbar_step, Pbar_step)

        dV = cls._deltaV(T_K=T_K, P_bar=P_array, melt_Fe=melt_Fe) * 1e-6  # cm3 to m3

        # TODO  CHECK UNITS: CM3, M3, A3, BAR, PASCAL OR GPA??
        return np.trapz(dV, P_array * 1e5)  # bar to Pascal

    @classmethod
    def _Fe_activities(cls, melt_cation_fractions, T_K):
        """
        Calculate FeO1.5 and FeO activities in silicate melt. Supplementary Note 2
        """

        axis = [0, 1][isinstance(melt_cation_fractions, pd.DataFrame)]

        sum_margules = melt_cation_fractions.mul(cls.margules).sum(axis=axis)

        LN_aFe3_aFe2 = sum_margules / (R * T_K) + (
            melt_cation_fractions["Fe"] - melt_cation_fractions["Fe3"]
        ) * cls.Fe_margules / (R * T_K)

        return LN_aFe3_aFe2

    @classmethod
    def get_error(cls, Fe3Fe2, pressure: Optional[pd.Series] = None, *args, **kwargs):

        return super().get_error(
            Fe3Fe2=Fe3Fe2,
            error_params_1bar=error_params_1bar,
            error_params_high_pressure=error_params_high_pressure,
            pressure=pressure,
        )


class Oneill2006(Fe3Fe2_model):
    """
    O'neill et al. (2006)\ [6]_
    """

    components = ["MgO", "CaO", "Na2O", "K2O", "Al2O3", "P2O5", "FeO"]

    error_params = [2.91079515e-01, -1.37913265e-02, 9.82946462e-01, 1.90242796e02]

    @classmethod
    def calculate_Fe3Fe2(
        cls,
        melt_mol_fractions,
        P_bar,
        T_K,
        fO2,
        Fe3Fe2_init=0.3,
        total_Fe="FeO",
        *args,
        **kwargs,
    ):
        """
        Calculate melt |Fe3Fe2| ratios with equation 10.

        Parameters
        ----------
        melt_mol_fractions   :   :py:class:`Pandas DataFrame <pandas:pandas.DataFrame>`
            Melt composition in oxide mol fractions
        P_bar : float, array-like
            pressure in bar
        T_K :   float, array-like
            temperature in Kelvin
        fO2 :   float, array-like
            Oxygen fugacity

        Returns
        -------
        float, array-like
            melt |Fe3Fe2| ratio
        """

        moles = (
            melt_mol_fractions.to_frame().T
            if isinstance(melt_mol_fractions, pd.Series)
            else melt_mol_fractions.copy()
        )  # force to DataFrame

        try:
            # extend the moles dataframe is multiple P,T, fO2 conditions are given for a single composition.
            if len(moles) != (l := len(T_K)):
                moles = moles.loc[moles.index.repeat(l)].reset_index(drop=True)
        except TypeError:
            pass
        # force everything into an iterable
        P_bar, T_K, fO2 = make_equal_length(P_bar, T_K, fO2)

        Fe3Fe2_func = (
            lambda Fe3Fe2_guess, moles, P, T, fO2: cls._calculate_Fe3Fe2(
                melt_mol_fractions=moles,
                P_bar=P,
                T_K=T,
                fO2=fO2,
                Fe3Fe2=Fe3Fe2_guess[0],
                total_Fe=total_Fe,
            )
            - Fe3Fe2_guess[0]
        )

        Fe3Fe2 = pd.Series(dtype=float)

        for (n, X), P, T, f in zip(moles.iterrows(), P_bar, T_K, fO2):
            Fe3Fe2.loc[n] = fsolve(Fe3Fe2_func, x0=Fe3Fe2_init, args=(X, P, T, f))[0]

        return Fe3Fe2.squeeze()

    @classmethod
    def _calculate_Fe3Fe2(
        cls, melt_mol_fractions, P_bar, T_K, fO2, Fe3Fe2=0.3, total_Fe: str = "FeO"
    ):

        P_GPa = P_bar / 1e4
        moles = check_components(
            composition=melt_mol_fractions, components=cls.components
        )
        cations = moles.FeO_Fe2O3_calc(Fe3Fe2, wtpc=False, total_Fe=total_Fe).cations()

        part_1 = (
            -28144
            + 3905 * cations["Mg"]
            - 13359 * cations["Ca"]
            - 14858 * cations["Na"]
            - 9805 * cations["K"]
            + 10906 * cations["Al"]
            + 110971 * cations["P"]
            - 11952 * (cations["Fe"] - cations["Fe3"])
        ) / T_K

        part_2 = (
            13.95
            + (33122 / T_K - 5.24) * ((1 + 0.241 * P_GPa) ** (3 / 4) - 1)
            - (39156 / T_K - 6.17) * ((1 + 0.132 * P_GPa) ** (3 / 4) - 1)
        )

        return 10 ** ((np.log10(fO2) - part_1 - part_2) / 4)

    @classmethod
    def get_error(cls, Fe3Fe2, pressure: Optional[pd.Series] = None, *args, **kwargs):

        return super().get_error(
            Fe3Fe2=Fe3Fe2,
            error_params_1bar=error_params_1bar,
            error_params_high_pressure=error_params_high_pressure,
            pressure=pressure,
        )


class Oneill2018(Fe3Fe2_model):
    """
    O'Neill et al. (2018)\ [7]_
    """

    components = ["CaO", "Na2O", "K2O", "P2O5"]

    error_params = [-1.99768682e-02, 7.41622563e-02, 9.80282471e-01, 1.28905955e02]

    @classmethod
    def calculate_Fe3Fe2(cls, melt_mol_fractions, T_K, fO2, *args, **kwargs):
        """
        Calculate melt |Fe3Fe2| ratios with equation 9a.

        Parameters
        ----------
        melt_mol_fractions   :   :py:class:`Pandas DataFrame <pandas:pandas.DataFrame>`
            Melt composition in oxide mol fractions
        T_K :   float, array-like
            temperature in Kelvin
        fO2 :   float, array-like
            Oxygen fugacity in bar

        Returns
        -------
        float, array-like
            melt |Fe3Fe2| ratio
        """
        cations = check_components(
            composition=melt_mol_fractions, components=cls.components
        ).cations()
        deltaQFM = cls._deltaQFM(fO2, T_K)

        return 10 ** (
            0.25 * deltaQFM
            - 1.36
            + 2.4 * cations["Ca"]
            + 2.0 * cations["Na"]
            + 3.7 * cations["K"]
            - 2.4 * cations["P"]
        )

    @staticmethod
    def _deltaQFM(fO2, T_K):
        return np.log10(fO2) - (8.58 - 25050 / T_K)

    @classmethod
    def get_error(cls, Fe3Fe2, pressure: Optional[pd.Series] = None, *args, **kwargs):

        return super().get_error(
            Fe3Fe2=Fe3Fe2,
            error_params_1bar=error_params_1bar,
            error_params_high_pressure=error_params_high_pressure,
            pressure=pressure,
        )


class Armstrong2019(Fe3Fe2_model):
    """
    Armstrong et al. (2019)\ [8]_

    Calibrated with one andesitic and one MORB compositions + data from O'neill et al. (2006)\ [6]_ and Zhang et al. (2017)\ [9]_
    """

    components = ["MgO", "CaO", "Na2O", "K2O", "Al2O3"]

    eos_parameters = pd.DataFrame(
        {"K_0": [37, 12.6], "Kprime_0": [8, 1.3]}, index=["Fe2", "Fe3"]
    )
    # K_0: GPa, Kprime_0: GPa-1

    margules = pd.Series(
        {
            "Mg": -2248,
            "Ca": 7690,
            "Na": 8553,
            "K": 5644,
            "Al": -6278,
        }
    )  # Jayasuriya et al., (2004)
    Fe_margules = 6880

    error_params = [2.20466771e-01, 9.51885060e-03, 9.85463396e-01, 2.05555278e02]

    @classmethod
    def calculate_Fe3Fe2(
        cls,
        melt_mol_fractions,
        T_K,
        P_bar,
        fO2,
        Fe3Fe2_init=0.3,
        total_Fe="FeO",
        *args,
        **kwargs,
    ):
        """
        Calculate melt |Fe3Fe2| ratios.

        Parameters
        ----------
        melt_mol_fractions   :   :py:class:`Pandas DataFrame <pandas:pandas.DataFrame>`
            Melt composition in oxide mol fractions
        T_K :   float, array-like
            temperature in Kelvin
        P_bar : float, array-like
            pressure in bar
        fO2 :   float, array-like
            Oxygen fugacity

        Returns
        -------
        float, array-like
            melt |Fe3Fe2| ratio
        """

        moles = (
            melt_mol_fractions.to_frame().T
            if isinstance(melt_mol_fractions, pd.Series)
            else melt_mol_fractions.copy()
        )  # force to DataFrame
        # force everything into iterables
        P_bar, T_K, fO2 = make_equal_length(P_bar, T_K, fO2)

        try:
            # extend the moles dataframe is multiple P,T, fO2 conditions are given for a single composition.
            if len(moles) != (l := max([len(param) for param in (T_K, P_bar, fO2)])):
                moles = moles.loc[moles.index.repeat(l)].reset_index(drop=True)
        except TypeError:
            pass

        Fe3Fe2_func = (
            lambda Fe3Fe2_guess, moles, T, P, fO2: cls._calculate_Fe3Fe2(
                melt_mol_fractions=moles,
                T_K=T,
                P_bar=P,
                fO2=fO2,
                Fe3Fe2=Fe3Fe2_guess[0],
                total_Fe=total_Fe,
            )
            - Fe3Fe2_guess[0]
        )

        Fe3Fe2 = pd.Series(dtype=float)

        for (n, X), P, T, f in zip(moles.iterrows(), P_bar, T_K, fO2):
            Fe3Fe2.loc[n] = fsolve(Fe3Fe2_func, x0=Fe3Fe2_init, args=(X, T, P, f))[0]

        return Fe3Fe2.squeeze()

    @classmethod
    def _calculate_Fe3Fe2(
        cls,
        melt_mol_fractions,
        T_K,
        P_bar,
        fO2,
        Fe3Fe2=0.3,
        total_Fe: str = "FeO",
        *args,
        **kwargs,
    ):
        """
        Calculate melt |Fe3Fe2| ratios with Supplementary Materials eq. S12

        Parameters
        ----------
        melt_mol_fractions   :   :py:class:`Pandas DataFrame <pandas:pandas.DataFrame>`
            Melt composition in oxide mol fractions
        T_K :   float, array-like
            temperature in Kelvin
        P_bar : float, array-like
            pressure in bar
        fO2 :   float, array-like
            Oxygen fugacity

        Returns
        -------
        float, array-like
            melt |Fe3Fe2| ratio
        """

        moles = check_components(
            composition=melt_mol_fractions, components=cls.components
        )
        cations = moles.FeO_Fe2O3_calc(Fe3Fe2, wtpc=False, total_Fe=total_Fe).cations()

        gibbs0 = cls._Gibbs0(T_K=T_K)
        dVdP = cls._dVdP(P_bar=P_bar, T_K=T_K)
        Fe_activities = cls._Fe_activities(melt_cation_fractions=cations, T_K=T_K)

        return np.exp(np.log(fO2) / 4 - (gibbs0 + dVdP) / (R * T_K) + Fe_activities)

    @staticmethod
    def _Gibbs0(T_K):
        """
        Supplementary Materials equation S4
        """
        return -(16201 / T_K - 8.031) * (R * T_K)

    @classmethod
    def _dVdP(cls, P_bar, T_K, **kwargs):

        VdP_Fe3 = cls._VdP(P_bar=P_bar, T_K=T_K, phase="Fe3")
        VdP_Fe2 = cls._VdP(P_bar=P_bar, T_K=T_K, phase="Fe2")

        return VdP_Fe3 - VdP_Fe2

    @classmethod
    def _VdP(cls, P_bar, T_K, phase: str):
        """
        Supplementary Materials equation S10
        """
        params = cls.eos_parameters.loc[phase]
        P_GPa = P_bar / 1e4
        Kprime_prime_0 = (
            -params["Kprime_0"] / params["K_0"]
        )  # Supplementary Materials ref. 32

        V_0 = cls._V_0(T_K, phase=phase)  # mm3/mol

        # Supplementary Materials eqs. S7-S9
        a = (1 + params["Kprime_0"]) / (
            1 + params["Kprime_0"] + params["K_0"] * Kprime_prime_0
        )
        b = params["Kprime_0"] / params["K_0"] - (
            Kprime_prime_0 / (1 + params["Kprime_0"])
        )
        c = (1 + params["Kprime_0"] + params["K_0"] * Kprime_prime_0) / (
            params["Kprime_0"] ** 2
            + params["Kprime_0"]
            - params["K_0"] * Kprime_prime_0
        )

        part_1 = a * (1 - (1 + b * P_GPa) ** (1 - c))
        part_2 = b * (c - 1) * P_GPa

        # TODO CHECK UNITS M3, CM3, BAR, PASCAL? -> M3 * (pressure, GPa?)
        return P_GPa * V_0 * (1 - a + part_1 / part_2)  # mm3.GPa = m3.Pa

    @staticmethod
    def _V_0(T_K, phase: str):
        """
        Partial molar volume of FeO or FeO1.5 in mm3/mol.

        Supplementary Materials ref 60, 61
        Lange & Carmichael (1987) values from table 8
        Lange & Carmichael (1990)
        """
        # TODO CHECK UNITS
        return {
            "Fe2": (13650 + 2.92 * (T_K - 1673)),
            "Fe3": (21070 + 4.54 * (T_K - 1673)),
        }[phase]

    @classmethod
    def _Fe_activities(cls, melt_cation_fractions, T_K):
        """
        Calculate FeO1.5 and FeO activities in silicate melt. Supplementary Materials eq. S11
        """

        axis = [0, 1][isinstance(melt_cation_fractions, pd.DataFrame)]

        sum_margules = melt_cation_fractions.mul(cls.margules).sum(axis=axis)

        LN_aFe3_aFe2 = sum_margules / T_K + cls.Fe_margules * (
            (melt_cation_fractions["Fe"] - melt_cation_fractions["Fe3"]) / T_K
        )

        return LN_aFe3_aFe2

    @classmethod
    def get_error(cls, Fe3Fe2, pressure: Optional[pd.Series] = None, *args, **kwargs):

        return super().get_error(
            Fe3Fe2=Fe3Fe2,
            error_params_1bar=error_params_1bar,
            error_params_high_pressure=error_params_high_pressure,
            pressure=pressure,
        )


class Zhang2017(Fe3Fe2_model):
    """
    Zhang et al. (2017)\ [9]_

    Only calibrated with an andesitic melt composition
    """

    parameters = pd.DataFrame(
        {
            "dVdT": [2.92, 3.69],
            "a": [-6.376, -6.627],
            "b": [107257, 110729],
            "c": [15095, 15243],
            "d": [8.27e-2, 1.137e-1],
            "K_0": [36.61, 27.11],
        },
        index=["LC", "Guo"],
    )  # dVdT: J/GPa/K

    error_params = [1.55144202e-01, 3.39642134e-03, 9.87246138e-01, 2.68123961e02]

    @classmethod
    def calculate_Fe3Fe2(
        cls, melt_mol_fractions, T_K, P_bar, fO2, parameters="LC", *args, **kwargs
    ):
        """
        Calculate melt |Fe3Fe2| ratios with equation 11.

        Parameters
        ----------
        melt_mol_fractions   :   :py:class:`Pandas DataFrame <pandas:pandas.DataFrame>`
            Melt composition in oxide mol fractions
        T_K :   float, array-like
            temperature in Kelvin
        P_bar : float, array-like
            pressure in bar
        fO2 :   float, array-like
            Oxygen fugacity

        Returns
        -------
        float, array-like
            melt |Fe3Fe2| ratio
        """
        params = cls.parameters.loc[parameters]
        d = params["d"]  # 4 / params["K_0"]

        # P_bar, T_K = make_iterable(P_bar, T_K)
        # array_length_one = np.array([len(T_K), len(P_bar)]) == 1

        # if array_length_one.any():
        #     array_length = np.array([len(T_K), len(P_bar)]).max()
        #     P_bar, T_K = [
        #         np.ones(array_length) * i[0] if len(i) == 1 else i for i in (P_bar, T_K)
        #     ]

        P_GPa = P_bar * 1e5 / 1e9

        part_1 = np.log(fO2) / 4 + params["a"] + params["b"] / (R * T_K)
        part_2 = (
            -(20170 + 4.54 * (T_K - 1673))
            * (16.6 / 3)
            * ((1 + 0.241 * P_GPa) ** (0.75) - 1)
            / (R * T_K)
        )
        part_3 = (params["c"] + params["dVdT"] * (T_K - 1673)) * (4 / (3 * d))
        part_4 = ((1 + d * P_GPa) ** (0.75) - 1) / (R * T_K)

        LN_Fe3Fe2 = part_1 + part_2 + part_3 * part_4

        return np.exp(LN_Fe3Fe2)

    @classmethod
    def get_error(cls, Fe3Fe2, pressure: Optional[pd.Series] = None, *args, **kwargs):

        return super().get_error(
            Fe3Fe2=Fe3Fe2,
            error_params_1bar=error_params_1bar,
            error_params_high_pressure=error_params_high_pressure,
            pressure=pressure,
        )


class Hirschmann2022(Fe3Fe2_model):
    """
    Hirschmann (2022)\ [10]_
    """

    components = ["SiO2", "TiO2", "MgO", "CaO", "Na2O", "K2O", "P2O5", "Al2O3"]

    params = pd.Series(
        {
            "a": 0.1917,
            "b": -1.961,
            "c": 4158.1,
            "dCp": 33.25,
            "T0": 1673.15,
            "y1": -520.46,
            "y2": -185.37,
            "y3": 494.39,
            "y4": 1838.34,
            "y5": 2888.48,
            "y6": 3473.68,
            "y7": -4473.6,
            "y8": -1245.09,
            "y9": -1156.86,
        }
    )
    # TODO calculate parameters
    error_params = [1, 1, 1, 1]

    @classmethod
    def calculate_Fe3Fe2(
        cls,
        melt_mol_fractions,
        T_K,
        P_bar,
        fO2,
        dVdP_method="Armstrong2019",
        *args,
        **kwargs,
    ):
        """
        Calculate melt |Fe3Fe2| ratios with equation 21.

        Parameters
        ----------
        melt_mol_fractions   :   :py:class:`Pandas DataFrame <pandas:pandas.DataFrame>`
            Melt composition in oxide mol fractions
        T_K :   float, array-like
            temperature in Kelvin
        P_bar : float, array-like
            pressure in bar
        fO2 :   float, array-like
            Oxygen fugacity

        Returns
        -------
        float, array-like
            melt |Fe3Fe2| ratio
        """
        moles = check_components(
            composition=melt_mol_fractions, components=cls.components
        )
        cations = moles.cations()

        compositional_term = cls._compositional_term(cations=cations)
        dVdP = cls._dVdP(T_K=T_K, P_bar=P_bar, method=dVdP_method)

        part_1 = (
            cls.params["a"] * np.log10(fO2) + cls.params["b"] + cls.params["c"] / T_K
        )
        part_2 = (
            -cls.params["dCp"]
            / (R * np.log(10))
            * (1 - cls.params["T0"] / T_K - np.log(T_K / cls.params["T0"]))
        )
        part_3 = -dVdP / (R * T_K * np.log(10))
        part_4 = 1 / T_K * compositional_term

        return 10 ** (part_1 + part_2 + part_3 + part_4)

    @staticmethod
    def _dVdP(T_K, P_bar, method="Armstrong2019"):

        model = getattr(sys.modules[__name__], method)

        return model._dVdP(T_K=T_K, P_bar=P_bar, melt_Fe="12.5molpc")

    @classmethod
    def _compositional_term(cls, cations):

        axis = [0, 1][isinstance(cations, pd.DataFrame)]

        part_1 = (
            cations[["Si", "Ti", "Mg", "Ca", "Na", "K", "P"]]
            .mul(cls.params[["y1", "y2", "y3", "y4", "y5", "y6", "y7"]].values)
            .sum(axis=axis)
        )
        part_2 = (
            cls.params["y8"] * cations["Si"] * cations["Al"]
            + cls.params["y9"] * cations["Si"] * cations["Mg"]
        )

        return part_1 + part_2

    @classmethod
    def get_error(cls, Fe3Fe2, pressure: Optional[pd.Series] = None, *args, **kwargs):

        return super().get_error(
            Fe3Fe2=Fe3Fe2,
            error_params_1bar=error_params_1bar,
            error_params_high_pressure=error_params_high_pressure,
            pressure=pressure,
        )


class Sun2024(Fe3Fe2_model):
    """
    Sun and Yao (2024)\ [11]_
    """

    components = ["FeO", "SiO2", "Al2O3", "TiO2", "CaO", "MgO", "Na2O", "K2O"]

    params = pd.Series(
        {
            "a0": 2.1479,
            "a1": -230.2593,
            "a2": -1.8557e-4,
            "a3": 34.3293,
            "a4": 1.4138,
            "a5": -17.3040,
            "a6": -10.1820,
            "a7": -6.7463,
            "a8": -7.3886,
            "a9": -14.5430,
            "a10": -9.9776,
            "a11": -16.1506,
            "a12": -37.5572,
            "h": 2.1410,
        }
    )
    # TODO calculate parameters
    error_params = [1, 1, 1, 1]

    @classmethod
    def _Gamma(
        cls,
        T_K: float | np.ndarray,
        P_bar: float | np.ndarray,
        **kwargs,
    ):
        """
        Calculate dV according to Deng (2020)
        """
        return Deng2020._dVdP(T_K=T_K, P_bar=P_bar, melt_Fe="12.5molpc", **kwargs) / (
            R * T_K
        )

    @classmethod
    def _Phi(cls, cations):

        axis = [0, 1][isinstance(cations, pd.DataFrame)]

        return (
            cls.params["a4"] * np.log(cations["Fe"])
            + cls.params["a5"] * cations["Fe"] ** 0.5
            + cls.params["a6"] * cations["Si"] ** 3.0
            + cations[["Al", "Ti", "Ca", "Mg"]]
            .mul(cls.params[["a7", "a8", "a9", "a10"]].values)
            .sum(axis=axis)
            + (cls.params["a11"] + cls.params["a12"] * cations["Fe"])
            * (cations["Na"] + cations["K"])
        )

    @classmethod
    def _Omega(cls, T_K):

        return (
            cls.params["a1"]
            + cls.params["a2"] * T_K**1.5
            + cls.params["a3"] * np.log(T_K)
        )

    @classmethod
    def calculate_Fe3Fe2(cls, melt_mol_fractions, T_K, P_bar, fO2, *args, **kwargs):
        """
        Calculate melt |Fe3Fe2| ratios with equation 9.

        Parameters
        ----------
        melt_mol_fractions   :   :py:class:`Pandas DataFrame <pandas:pandas.DataFrame>`
            Melt composition in oxide mol fractions
        T_K :   float, array-like
            temperature in Kelvin
        P_bar : float, array-like
            pressure in bar
        fO2 :   float, array-like
            Oxygen fugacity
        Fe3Fe2 : float, array-like
            melt Fe3+/Fe2+ ratios

        Returns
        -------
        float, array-like
            melt |Fe3Fe2| ratio
        """
        moles = check_components(
            composition=melt_mol_fractions, components=cls.components
        )
        cations = moles.cations()

        omega = cls._Omega(T_K=T_K)
        phi = cls._Phi(cations=cations)
        gamma = cls._Gamma(T_K=T_K, P_bar=P_bar)

        return 10 ** (
            (np.log10(fO2) - omega - phi - cls.params["h"] * gamma)
            / (4 + cls.params["a0"] * cations["Fe"] ** 0.5)
        )

    @classmethod
    def get_error(cls, Fe3Fe2, pressure: Optional[pd.Series] = None, *args, **kwargs):

        return super().get_error(
            Fe3Fe2=Fe3Fe2,
            error_params_1bar=error_params_1bar,
            error_params_high_pressure=error_params_high_pressure,
            pressure=pressure,
        )


_clsmembers = inspect.getmembers(sys.modules[__name__], inspect.isclass)
# Collect all Fe3Fe2_models in a dictionary.
Fe3Fe2_models_dict = {cls[0]: cls[1] for cls in _clsmembers if _is_Fe3Fe2_model(cls[1])}



# configuration needs to be imported first to avoid circular imports
from MagmaPandas.configuration import *  # isort: skip

from MagmaPandas.MagmaFrames import *
from MagmaPandas.MagmaSeries import *
from MagmaPandas.parse_io.readers import *



from MagmaPandas.configuration import configuration
from MagmaPandas.Fe_redox.Fe3Fe2_models import Fe3Fe2_models_dict
from MagmaPandas.fO2 import fO2_calculate as f


def calculate_Fe3Fe2(mol_fractions, T_K, P_bar, fO2=None, **kwargs):
    """
    Calculate melt |Fe3Fe2| with the configured *f*\ O2 buffer and |Fe3Fe2| model.

    Model choices are set in the global :py:class:`~MagmaPandas.configuration.configuration` class.

    Parameters
    ----------
    mol_fractions   :   :py:class:`Pandas DataFrame <pandas:pandas.DataFrame>`
        Melt composition in oxide mol_fractions
    T_K  :       float, pd.Series-like
        temperature in Kelvin
    P_bar   :     float, pd.Series-like
        Pressure in bars
    logshift  :  int, pd.Series-like
        log units shift of QFM buffer
    model     :  string
        'kressCarmichael' or 'borisov'

    Returns
    -------
    melt Fe\ :sup:`3+`\ /Fe\ :sup:`2+` ratio
    """

    Fe_model_name = kwargs.get("Fe_model", configuration.Fe3Fe2_model)
    Fe_model = Fe3Fe2_models_dict[Fe_model_name]  # getattr(Fe_redox, Fe_model_name)

    if fO2 is None:
        fO2 = f.calculate_fO2(T_K=T_K, P_bar=P_bar, **kwargs)

    return Fe_model.calculate_Fe3Fe2(
        melt_mol_fractions=mol_fractions, T_K=T_K, P_bar=P_bar, fO2=fO2
    )



from abc import ABC, abstractmethod
from typing import Dict, Optional

import numpy as np
import pandas as pd
from scipy import interpolate

# from MagmaPandas.Fe_redox.Fe3Fe2_errors import (
#     error_params_1bar,
#     error_params_high_pressure,
# )
from MagmaPandas.tools.model_errors import _error_func

# Fe3+/Fe2+ limits of the moving standard deviation in a 30 point window of the validation dataset (provided at ./data/Fe3Fe2_validation_data.csv).
validation_limits_1bar = (0.0351966873706004, 5.948890681577911)
validation_limits_high_pressure = (0.052631579, 2.160641174)


def _Fe3Fe2_error_func(a, b, c, d, Fe3Fe2):
    return _error_func(data=Fe3Fe2, a=a, b=b, c=c, d=d)


def _Fe3Fe2_spline(Fe3Fe2, parameters):
    return interpolate.splev(Fe3Fe2, parameters)


class Fe3Fe2_model(ABC):
    @abstractmethod
    def calculate_Fe3Fe2(
        cls, melt_mol_fractions, T_K, fO2, *args, **kwargs
    ) -> float | np.ndarray:
        """
        Calculate melt |Fe3Fe2| ratios

        Parameters
        ----------
        melt_mol_fractions   :   :py:class:`Pandas DataFrame <pandas:pandas.DataFrame>`
            Melt composition in oxide mol fractions
        T_K :   float, array-like
            temperature in Kelvin
        fO2 :   float, array-like
            Oxygen fugacity

        Returns
        -------
        float, array-like
            melt |Fe3Fe2| ratio
        """
        pass

    @classmethod
    def _calculate_Fe3Fe2_(
        cls, melt_mol_fractions, T_K, fO2, offset_parameters=0.0, *args, **kwargs
    ):
        """
        Calculate melt Fe3Fe2 ratios and offset results by ``offset_paramaters`` * standard deviation of the model.
        """
        Fe3Fe2 = cls.calculate_Fe3Fe2(
            melt_mol_fractions=melt_mol_fractions,
            T_K=T_K,
            fO2=fO2,
            **kwargs,
        )

        if offset_parameters == 0.0:
            return Fe3Fe2

        # if self._Fe3Fe2_offset_parameters != 0.0:
        offset = cls.get_offset(
            melt_composition=melt_mol_fractions,
            Fe3Fe2=Fe3Fe2,
            offset_parameters=offset_parameters,
            pressure=kwargs["P_bar"],
        )

        Fe3Fe2 = Fe3Fe2 + offset

        # Make sure no negative values are returned
        try:
            Fe3Fe2[Fe3Fe2 <= 0.0] = 1e-6
        except TypeError:
            Fe3Fe2 = 1e-6 if Fe3Fe2 <= 0 else Fe3Fe2

        # Make sure arrays of length 1 are returned as floats instead
        try:
            if len(Fe3Fe2) == 1:
                return np.array(Fe3Fe2)[0]
            return Fe3Fe2

        except TypeError:
            return Fe3Fe2

    @classmethod
    def get_error(
        cls,
        Fe3Fe2,
        error_params_1bar: Dict,
        error_params_high_pressure: Dict,
        pressure: Optional[pd.Series] = None,
        *args,
        **kwargs,
    ) -> float | np.ndarray:
        """
        Returns one standard deviation error on |Fe3Fe2| ratios, calculated from a compiled validation dataset.

        Parameters
        ----------
        Fe3Fe2  : array-like
            melt |Fe3Fe2| ratios
        pressure : array-like, optional
            pressures of each element in ``|Fe3Fe2|``. If this term is not included, errors will be calculated strictly at 1 bar.

        Returns
        -------
        float, array-like
            |Fe3Fe2| error
        """

        name = cls.__name__
        # TODO change error_params_1bar to a method argument to reduce dependencies
        error_1bar = _Fe3Fe2_error_func(*error_params_1bar[name], Fe3Fe2=Fe3Fe2)

        if pressure is None:
            return error_1bar

        # TODO change error_params_high_pressure to a method argument to reduce dependencies
        error_high_pressure = _Fe3Fe2_spline(
            Fe3Fe2=Fe3Fe2, parameters=error_params_high_pressure[name]
        )

        errors = error_1bar.copy()

        # catch index errors for ints, floats and 0-dimensional arrays.
        try:
            error_high_pressure[0]
        except (TypeError, IndexError):
            # convert everyting to a 1-dimensional arrays.
            pressure = np.array([pressure]).flatten()
            errors = np.array([errors]).flatten()
            error_high_pressure = np.array([error_high_pressure]).flatten()

        hp = pressure > 1
        errors[hp] = error_high_pressure[hp]

        return errors

    @classmethod
    def get_offset(
        cls, Fe3Fe2, offset_parameters, pressure=None, *args, **kwargs
    ) -> float | np.ndarray:

        return cls.get_error(Fe3Fe2=Fe3Fe2, pressure=pressure) * offset_parameters

    @staticmethod
    def get_offset_parameters(n: int = 1) -> float | np.ndarray:
        return np.random.normal(loc=0, scale=1, size=n)



import numpy as np
import pandas as pd
from scipy.constants import R

"""
Based on supplementary text to:

K. Armstrong et al. (2019) Deep magma ocean formation set the oxidation state of Earth’s mantle. Science. Vol 365, pp. 903-906

unless specified otherwise
"""

Fe_bulk_modulus = pd.DataFrame(
    {"k0": [45.8, 37], "k0_p": [4.7, 8]}, index=("Fe_metal", "FeO_melt")
)  # GPa

activities = pd.Series({"FeO_melt": 1.5, "Fe_metal": 1})


def Gibbs0(T_K):
    """
    Standard state free energy of the reaction

    Fe(metal) + 0.5O2 -> FeO(liq)

    Table 6 from:

    St.C. O'Neill et al. (2002) The effect of melt composition on trace element partitioning: an experimental investigation of the activity coefficients of FeO, NiO, CoO, MoO2 and MoO3 in silicate melts. Chemical Geology. Vol 186, pp. 151-181

    J/mol
    """

    return -244118 + 115.559 * T_K - 8.474 * T_K * np.log(T_K)


def V0_FeO(T_K):
    """
    Equation S10

    J/GPa
    """

    return 13650 + 2.92 * (T_K - 1673)


def V0_Feliquid(T_K):
    """
    page 10
    Why is this the same as V0(FeO)?
    THIS IS WRONG

    J/GPa
    """
    return 13650 + 2.92 * (T_K - 1673)


def _a(k0, k0_p):
    """
    S7
    """
    k0_pp = -k0_p / k0
    return (1 + k0_p) / (1 + k0_p + k0 * k0_pp)


def _b(k0, k0_p):
    """
    S8
    """
    k0_pp = -k0_p / k0
    return k0_p / k0 - k0_pp / (1 + k0_p)


def _c(k0, k0_p):
    """
    S9
    """
    k0_pp = -k0_p / k0
    return (1 + k0_p + k0 * k0_pp) / (k0_p**2 + k0_p - k0 * k0_pp)


def VdP(P_bar, T_K, phase):
    """
    S10
    """

    P_GPa = P_bar / 1e4

    k0, k0_p = Fe_bulk_modulus.loc[phase]

    V0 = {"Fe_metal": V0_Feliquid, "FeO_melt": V0_FeO}[phase](T_K=T_K)

    a = _a(k0=k0, k0_p=k0_p)
    b = _b(k0=k0, k0_p=k0_p)
    c = _c(k0=k0, k0_p=k0_p)

    part_1 = a * (1 - (1 + b + P_GPa) ** (1 - c))
    part_2 = b * (c - 1) * P_GPa

    return P_GPa * V0 * (1 - a + part_1 / part_2)


def deltaGibbs(P_bar, T_K):

    G0 = Gibbs0(T_K=T_K)
    VdP_FeO = VdP(P_bar=P_bar, T_K=T_K, phase="FeO_melt")
    VdP_Fe = VdP(P_bar=P_bar, T_K=T_K, phase="Fe_metal")
    dVdP = VdP_FeO - VdP_Fe

    return G0 + dVdP


def calculate_fO2(T_K, P_bar, XFeO_melt, XFe_metal=1):
    """
    S14

    calculate fO2 at the equilibrium

    2Fe(liquid) + O2 -> 2FeO(melt)

    based om measured melt FeO mol fractions.
    """

    dG = deltaGibbs(P_bar=P_bar, T_K=T_K)

    log10fO2 = (
        dG / (R * T_K * np.log(10))
        + 2 * np.log10(XFeO_melt * activities["FeO_melt"])
        - 2 * np.log10(XFe_metal * activities["Fe_metal"])
    )

    return 10**log10fO2
